from typing import List, Dict

import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class LongitudinalClassifierConfig:
    """Manages the configuration space for classifiers

    Focus: longitudinal supervised learning contexts

    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that defines the hyperparameters and their ranges for
        the classifiers.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if any(
            meta_key not in config_space.meta
            for meta_key in ["data_preparation", "estimators"]
        ):
            raise ValueError(
                f"Expected 'data_preparation', 'estimators', "
                f"keys in meta of config_space. "
                f"Got {config_space.meta.keys()}"
            )
        if (
            config_space.get_hyperparameter(config_space.meta["data_preparation"])
            is None
        ):
            raise ValueError(
                f"Expected '{config_space.meta['data_preparation']}' hyperparameter "
                f"in config_space"
            )
        self.config_space = config_space
        self.classifiers_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            # Longitudinal Algorithms
            "NestedTreesClassifier": self.setup_nested_trees,
            "LexicoRandomForestClassifier": self.setup_lexico_rf_classifier,
            "LexicoDecisionTreeClassifier": self.setup_lexico_decision_tree_classifier,
            "LexicoDeepForestClassifier": self.setup_lexico_deep_forest_classifier,
            "LexicoGradientBoostingClassifier": self.setup_lexico_gradient_boosting,
            # Conventional Algorithms
            "DecisionTreeClassifier": self.setup_decision_tree,
            "RandomForestClassifier": self.setup_random_forest,
            "ExtraTreesClassifier": self.setup_extra_trees,
            "KNeighborsClassifier": self.setup_k_neighbors,
            "LinearSVC": self.setup_linear_svc,
            "DeepForestClassifier": self.setup_deep_forest,
            "GradientBoostingClassifier": self.setup_gradient_boosting,
        }
        self.forbidden_clauses: List[Dict[str, List[str]]] = [
            {
                "data_preparations": ["MerWavTimePlus"],
                "classifiers_to_forbid": [
                    "DecisionTreeClassifier",
                    "RandomForestClassifier",
                    "ExtraTreesClassifier",
                    "KNeighborsClassifier",
                    "LinearSVC",
                    "DeepForestClassifier",
                    "GradientBoostingClassifier",
                ],
            },
            {
                "data_preparations": [
                    "MerWavTimeMinus",
                    "AggrFuncMean",
                    "AggrFuncMedian",
                    # "AggrFuncMode",
                    "SepWavMajorityVoting",
                    "SepWavDecayLinearWeighting",
                    "SepWavDecayExponentialWeighting",
                    "SepWavDecayCrossValidation",
                    "SepWavStackingLogisticRegression",
                    "SepWavStackingDecisionTree",
                    "SepWavStackingRandomForest",
                ],
                "classifiers_to_forbid": [
                    "NestedTreesClassifier",
                    "LexicoRandomForestClassifier",
                    "LexicoDecisionTreeClassifier",
                    "LexicoDeepForestClassifier",
                    "LexicoGradientBoostingClassifier",
                ],
            },
        ]
        self.cs_data_preparation_name = self.config_space.meta["data_preparation"]
        self.cs_estimators_name = self.config_space.meta["estimators"]

    @property
    def shared_hyperparameters(self):
        return {
            "criterion": ["gini", "entropy"],
            "max_depth": {"lower": 2, "upper": 10, "default_value": 2},
            "min_samples_split": {"lower": 2, "upper": 20, "default_value": 2},
            "min_samples_leaf": {"lower": 1, "upper": 20, "default_value": 1},
            "n_estimators": {"lower": 100, "upper": 1000, "default_value": 100},
            "n_estimators_deep_forests": {"lower": 2, "upper": 3, "default_value": 2},
            "bootstrap": [True, False],
            "threshold_gain": {
                "choices": [
                    0.0,
                    0.001,
                    0.0015,
                    0.002,
                    0.0025,
                    0.003,
                    0.0035,
                    0.004,
                    0.0045,
                    0.005,
                    0.01,
                ],
                "default_value": 0.0,
            },
        }

    def setup_classifiers(self):
        classifiers_choices = list(self.classifiers_setup_map.keys())

        if not classifiers_choices:
            raise ValueError("No classifiers to add to config space")

        classifiers = csh.CategoricalHyperparameter(
            name=self.cs_estimators_name,
            choices=classifiers_choices,
            default_value=classifiers_choices[0],
        )
        self.config_space.add_hyperparameter(classifiers)

        if self.forbidden_clauses is not None:
            if (
                self.cs_data_preparation_name
                not in self.config_space.get_hyperparameter_names()
            ):
                raise ValueError(
                    f"Data preparation hyperparameter "
                    f"{self.cs_data_preparation_name} not found in config space"
                )
            data_preparation = self.config_space.get_hyperparameter(
                self.cs_data_preparation_name
            )
            data_preparation_choices = list(data_preparation.choices)
            forbidden_clauses = [
                cs.ForbiddenAndConjunction(
                    cs.ForbiddenInClause(classifiers, clause["classifiers_to_forbid"]),
                    cs.ForbiddenInClause(data_preparation, clause["data_preparations"]),
                )
                for clause in self.forbidden_clauses
                if all(
                    data_preparation_name in data_preparation_choices
                    for data_preparation_name in clause["data_preparations"]
                )
            ]
            self.config_space.add_forbidden_clauses(forbidden_clauses)

        for classifier_name in classifiers_choices:
            if setup_func := self.classifiers_setup_map.get(classifier_name):
                setup_func(classifiers)

    def _add_hyperparameters_and_equals_conditions(
        self,
        local_vars: dict,
        estimator_name: str,
    ):
        if "classifiers" not in local_vars or not isinstance(
            local_vars["classifiers"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'classifiers' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["classifiers"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["classifiers"], estimator_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_nested_trees(self, classifiers: csh.CategoricalHyperparameter):
        max_outer_depth = csh.UniformIntegerHyperparameter(
            "max_outer_depth__NestedTreesClassifier",
            lower=5,
            upper=10,
            default_value=10,
        )
        max_inner_depth = csh.UniformIntegerHyperparameter(
            "max_inner_depth__NestedTreesClassifier",
            lower=2,
            upper=5,
            default_value=5,
        )
        min_node_size = csh.UniformIntegerHyperparameter(
            "min_outer_samples__NestedTreesClassifier",
            lower=2,
            upper=10,
            default_value=2,
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "NestedTreesClassifier"
        )

    def setup_lexico_rf_classifier(self, classifiers: csh.CategoricalHyperparameter):
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__LexicoRandomForestClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__LexicoRandomForestClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__LexicoRandomForestClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__LexicoRandomForestClassifier",
            self.shared_hyperparameters["bootstrap"],
        )
        threshold_gain = csh.CategoricalHyperparameter(
            "threshold_gain__LexicoRandomForestClassifier",
            **self.shared_hyperparameters["threshold_gain"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "LexicoRandomForestClassifier"
        )

    def setup_lexico_decision_tree_classifier(
        self, classifiers: csh.CategoricalHyperparameter
    ):
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__LexicoDecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__LexicoDecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__LexicoDecisionTreeClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        threshold_gain = csh.CategoricalHyperparameter(
            "threshold_gain__LexicoDecisionTreeClassifier",
            **self.shared_hyperparameters["threshold_gain"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "LexicoDecisionTreeClassifier"
        )

    def setup_lexico_deep_forest_classifier(
        self, classifiers: csh.CategoricalHyperparameter
    ):
        single_classifier_type = csh.CategoricalHyperparameter(
            "single_classifier_type__LexicoDeepForestClassifier",
            ["LexicoRandomForestClassifier", "LexicoCompleteRFClassifier"],
        )
        single_count = csh.UniformIntegerHyperparameter(
            "single_count__LexicoDeepForestClassifier",
            **self.shared_hyperparameters["n_estimators_deep_forests"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "LexicoDeepForestClassifier"
        )

    def setup_decision_tree(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__DecisionTreeClassifier",
            self.shared_hyperparameters["criterion"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__DecisionTreeClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "DecisionTreeClassifier"
        )

    def setup_random_forest(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__RandomForestClassifier",
            self.shared_hyperparameters["criterion"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__RandomForestClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__RandomForestClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__RandomForestClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__RandomForestClassifier",
            self.shared_hyperparameters["bootstrap"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "RandomForestClassifier"
        )

    def setup_extra_trees(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__ExtraTreesClassifier",
            self.shared_hyperparameters["criterion"],
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__ExtraTreesClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__ExtraTreesClassifier",
            self.shared_hyperparameters["bootstrap"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "ExtraTreesClassifier"
        )

    def setup_k_neighbors(self, classifiers: csh.CategoricalHyperparameter):
        n_neighors = csh.CategoricalHyperparameter(
            "n_neighbors__KNeighborsClassifier",
            [1, 2, 3, 4, 5, 10, 15, 20, 30, 40, 50],
        )
        weights = csh.CategoricalHyperparameter(
            "weights__KNeighborsClassifier", ["uniform", "distance"]
        )
        p = csh.CategoricalHyperparameter("p__KNeighborsClassifier", [1, 2])
        self._add_hyperparameters_and_equals_conditions(
            locals(), "KNeighborsClassifier"
        )

    def setup_linear_svc(self, classifiers: csh.CategoricalHyperparameter):
        loss = csh.CategoricalHyperparameter(
            "loss__LinearSVC", ["hinge", "squared_hinge"]
        )
        penalty = csh.CategoricalHyperparameter("penalty__LinearSVC", ["l1", "l2"])
        dual = csh.CategoricalHyperparameter("dual__LinearSVC", [True, False])
        tol = csh.CategoricalHyperparameter(
            "tol__LinearSVC", [1e-5, 1e-4, 1e-3, 1e-2, 1e-1]
        )
        C = csh.CategoricalHyperparameter(
            "C__LinearSVC",
            [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "LinearSVC")

        # Forbidden clause: Penalty 'l1' cannot be used with loss 'hinge'
        forbidden_penalty_loss = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l1"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
        )
        # Forbidden clause: Penalty 'l2' cannot be used
        # with loss 'hinge' and dual 'False'
        forbidden_penalty_loss_dual_false = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l2"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
            cs.ForbiddenEqualsClause(self.config_space["dual__LinearSVC"], False),
        )
        self.config_space.add_forbidden_clauses(
            [forbidden_penalty_loss, forbidden_penalty_loss_dual_false]
        )

    def setup_deep_forest(self, classifiers: csh.CategoricalHyperparameter):
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__DeepForestClassifier",
            **self.shared_hyperparameters["n_estimators_deep_forests"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "DeepForestClassifier"
        )

    def setup_gradient_boosting(self, classifiers: csh.CategoricalHyperparameter):
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__GradientBoostingClassifier",
            lower=2,
            upper=10,
            default_value=2,
        )
        learning_rate = csh.UniformFloatHyperparameter(
            "learning_rate__GradientBoostingClassifier",
            lower=0.01,
            upper=0.5,
            default_value=0.1,
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__GradientBoostingClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "GradientBoostingClassifier"
        )

    def setup_lexico_gradient_boosting(
        self, classifiers: csh.CategoricalHyperparameter
    ):
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__LexicoGradientBoostingClassifier",
            lower=2,
            upper=10,
            default_value=2,
        )
        learning_rate = csh.UniformFloatHyperparameter(
            "learning_rate__LexicoGradientBoostingClassifier",
            lower=0.01,
            upper=0.5,
            default_value=0.1,
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__LexicoGradientBoostingClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        threshold_gain = csh.CategoricalHyperparameter(
            "threshold_gain__LexicoGradientBoostingClassifier",
            **self.shared_hyperparameters["threshold_gain"],
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "LexicoGradientBoostingClassifier"
        )



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class LongitudinalDataPreparationConfig:
    """Manages the configuration space for Data Preparation

    Focus: longitudinal supervised learning contexts

    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that defines the hyperparameters and their ranges for
        the classifiers.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "data_preparation" not in config_space.meta:
            raise ValueError(
                f"Expected 'data_preparation', "
                f"key in meta of config_space. "
                f"Got {config_space.meta.keys()}"
            )
        self.config_space = config_space
        self.data_preparation_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "MerWavTimePlus": self.setup_merwav_time_plus,
            "MerWavTimeMinus": self.setup_merwav_time_minus,
            "AggrFuncMean": self.setup_aggr_func_mean,
            "AggrFuncMedian": self.setup_aggr_func_median,
            # "AggrFuncMode": self.setup_aggr_func_mode,
            "SepWavMajorityVoting": self.setup_sep_wav_majority_voting,
            "SepWavDecayLinearWeighting": self.setup_sep_wav_decay_linear_weighting,
            "SepWavDecayExponentialWeighting": self.setup_sep_wav_decay_exponential_weighting,  # noqa: E501
            "SepWavDecayCrossValidation": self.setup_sep_wav_decay_cross_validation,
            "SepWavStackingLogisticRegression": self.setup_sep_wav_stacking_logistic_regression,  # noqa: E501
            "SepWavStackingDecisionTree": self.setup_sep_wav_stacking_decision_tree,
            "SepWavStackingRandomForest": self.setup_sep_wav_stacking_random_forest,
        }
        self.cs_data_preparation_name = self.config_space.meta["data_preparation"]

    def setup_data_preparation(self):
        data_preparation_choices = list(self.data_preparation_setup_map.keys())

        if not data_preparation_choices:
            raise ValueError("No data preparation to add to config space")

        weights = [
            (
                0.5
                if choice == "MerWavTimePlus"
                else 0.5 / (len(data_preparation_choices) - 1)
            )
            for choice in data_preparation_choices
        ]

        data_preparation = csh.CategoricalHyperparameter(
            name=self.cs_data_preparation_name,
            choices=data_preparation_choices,
            weights=weights,
            default_value=data_preparation_choices[0],
        )

        self.config_space.add_hyperparameter(data_preparation)

        for data_preparation_name in data_preparation_choices:
            if setup_func := self.data_preparation_setup_map.get(data_preparation_name):
                setup_func(data_preparation)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_merwav_time_plus(self, data_preparation: csh.CategoricalHyperparameter):
        pass

    def setup_merwav_time_minus(self, data_preparation: csh.CategoricalHyperparameter):
        pass

    def setup_aggr_func_mean(self, data_preparation: csh.CategoricalHyperparameter):
        pass

    def setup_aggr_func_median(self, data_preparation: csh.CategoricalHyperparameter):
        pass

    def setup_aggr_func_mode(self, data_preparation: csh.CategoricalHyperparameter):
        pass

    def setup_sep_wav_majority_voting(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_decay_linear_weighting(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_decay_exponential_weighting(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_decay_cross_validation(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_stacking_logistic_regression(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_stacking_decision_tree(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass

    def setup_sep_wav_stacking_random_forest(
        self, data_preparation: csh.CategoricalHyperparameter
    ):
        pass



"""
Contains mutation functions for genetic programming.
Each mutation function takes an individual and modifies it in-place.
"""
import logging
import random
from functools import partial
from typing import Callable, Optional, cast, List, Dict

import ConfigSpace as cs
import numpy as np

from gama.genetic_programming.components import PrimitiveNode, Terminal
from .components import Individual, DATA_TERMINAL
from .operations import random_primitive_node
from ..utilities.config_space import get_internal_output_types

# Avoid stopit from logging warnings every time a pipeline evaluation times out
logging.getLogger("stopit").setLevel(logging.ERROR)
log = logging.getLogger(__name__)


def mut_replace_terminal(
    individual: Individual, config_space: cs.ConfigurationSpace
) -> None:
    """Mutates an Individual in-place by replacing one of its Terminals.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    config_space: ConfigurationSpace
        The ConfigSpace object which defines the search space. Refer to the
        configuration/(classification||regression).py file for further details.
    """

    def terminal_replaceable(index_terminal):
        _, terminal = index_terminal
        return has_multiple_options(
            config_space.get_hyperparameter(terminal.config_space_name)
        )

    terminals = list(filter(terminal_replaceable, enumerate(individual.terminals)))
    if not terminals:
        raise ValueError("Individual has no terminals suitable for mutation.")

    terminal_index, old = random.choice(terminals)

    while True:
        new_terminal_value = config_space.get_hyperparameter(
            old.config_space_name
        ).sample(np.random.RandomState())
        if new_terminal_value != old.value:
            break

    new_terminal = Terminal(
        identifier=old.identifier,
        value=new_terminal_value,
        output=old.output,
        config_space_name=old.config_space_name,
    )
    individual.replace_terminal(terminal_index, new_terminal)


def mut_replace_primitive(
    individual: Individual, config_space: cs.ConfigurationSpace
) -> None:
    """Mutates an Individual in-place by replacing one of its Primitives.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    config_space: cs.ConfigurationSpace
        The ConfigSpace object which defines the search space. Refer to the
        configuration/(classification||regression).py file for further details.
    """

    def primitive_replaceable(index_primitive):
        _, primitive = index_primitive
        return has_multiple_options(
            config_space.get_hyperparameter(
                config_space.meta[primitive._primitive.output]
                if primitive._primitive.output in get_internal_output_types()
                else primitive._primitive.output
            )
        )

    primitives = list(filter(primitive_replaceable, enumerate(individual.primitives)))
    if not primitives:
        raise ValueError("Individual has no primitives suitable for replacement.")

    primitive_index, old_primitive_node = random.choice(primitives)
    primitive_node = random_primitive_node(
        output_type=old_primitive_node._primitive.output,
        config_space=config_space,
        exclude=old_primitive_node._primitive,
    )
    individual.replace_primitive(primitive_index, primitive_node)


def mut_shrink(
    individual: Individual,
    _config_space: Optional[cs.ConfigurationSpace] = None,
    shrink_by: Optional[int] = None,
) -> None:
    """Mutates an Individual in-place by removing any number of primitive nodes.

    Primitive nodes are removed from the preprocessing end.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    _config_space: dict, optional
        Not used. Present to create a matching function signature with other mutations.
    shrink_by: int, optional (default=None)
        Number of primitives to remove.
        Must be at least one greater than the number of primitives in `individual`.
        If None, a random number of primitives is removed.
    """
    n_primitives = len(list(individual.primitives))
    if shrink_by is not None and n_primitives <= shrink_by:
        raise ValueError(f"Can't shrink size {n_primitives} individual by {shrink_by}.")
    if shrink_by is None:
        shrink_by = random.randint(1, n_primitives - 1)

    current_primitive_node = individual.main_node
    primitives_left = n_primitives - 1
    while primitives_left > shrink_by:
        current_primitive_node = cast(PrimitiveNode, current_primitive_node._data_node)
        primitives_left -= 1
    current_primitive_node._data_node = DATA_TERMINAL


def mut_insert(individual: Individual, config_space: cs.ConfigurationSpace) -> None:
    """Mutate an Individual in-place by inserting a PrimitiveNode at a random location.

    The new PrimitiveNode will not be inserted as root node.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    config_space: cs.ConfigurationSpace
        The ConfigSpace object which defines the search space. Refer to the
        configuration/(classification||regression).py file for further details.
    """
    parent_node = random.choice(list(individual.primitives))
    new_primitive_node = random_primitive_node(
        output_type=DATA_TERMINAL, config_space=config_space
    )
    new_primitive_node._data_node = parent_node._data_node
    parent_node._data_node = new_primitive_node


def has_multiple_options(hyperparameter: cs.hyperparameters.hyperparameter) -> bool:
    """Check if a ConfigSpace hyperparameter has more than one option.

    Only Constant, Float, Integer, and Categorical hyperparameters are currently
    supported. A TypeError is thrown if the hyperparameter is not of one of these
    types. Additionally, readers are directed to our Github Issues page to request
    support for other types.

    Parameters
    ----------
    hyperparameter: cs.hyperparameters.hyperparameter
        The hyperparameter to check.

    Returns
    -------
    bool
        True if the hyperparameter has more than one option, False otherwise.
    """
    if isinstance(
        hyperparameter,
        (
            cs.hyperparameters.FloatHyperparameter,
            cs.hyperparameters.IntegerHyperparameter,
        ),
    ):
        # For Float and Integer, check if the upper and lower bounds are different
        return hyperparameter.upper > hyperparameter.lower
    elif isinstance(hyperparameter, cs.CategoricalHyperparameter):
        # For Categorical, check if there are more than one unique items
        return len(set(hyperparameter.choices)) > 1
    elif isinstance(hyperparameter, cs.hyperparameters.Constant):
        # Constant has only one option
        return False
    else:
        # Default case, assuming no options or not a recognised type
        raise TypeError(f"Hyperparameter type {type(hyperparameter)} not supported")


def random_valid_mutation_in_place(
    individual: Individual,
    config_space: cs.ConfigurationSpace,
    max_length: Optional[int] = None,
) -> Callable:
    """Apply a random valid mutation in place.

    The random mutation can be one of:

     - mut_random_primitive
     - mut_random_terminal, if the individual has at least one
     - mutShrink, if individual has at least two primitives
     - mutInsert, if it would not exceed `new_max_length` when specified.

    Parameters
    ----------
    individual: Individual
        An individual to be mutated *in-place*.
    config_space: cs.ConfigurationSpace
        The ConfigSpace object which defines the search space. Refer to the
        configuration/(classification||regression).py file for further details.
    max_length: int, optional (default=None)
        If specified, impose a maximum length on the new individual.


    Returns
    -------
    Callable
        The mutation function used.
    """
    n_primitives = len(list(individual.primitives))
    available_mutations: List[Callable[[Individual, Dict], None]] = []
    if max_length is not None and n_primitives > max_length:
        available_mutations.append(
            partial(mut_shrink, shrink_by=n_primitives - max_length)
        )
    else:
        replaceable_primitives = filter(
            lambda p: has_multiple_options(
                config_space.get_hyperparameter(
                    config_space.meta[p._primitive.output]
                    if p._primitive.output in get_internal_output_types()
                    else p._primitive.output
                )
            ),
            individual.primitives,
        )

        if len(list(replaceable_primitives)) > 1:
            available_mutations.append(mut_replace_primitive)

        if max_length is None or n_primitives < max_length:
            available_mutations.append(mut_insert)
        if n_primitives > 1:
            available_mutations.append(mut_shrink)

        replaceable_terminals = filter(
            lambda t: has_multiple_options(
                config_space.get_hyperparameter(t.config_space_name)
            ),
            individual.terminals,
        )
        if len(list(replaceable_terminals)) > 1:
            available_mutations.append(mut_replace_terminal)

    if not available_mutations:
        log.warning(
            f"Individual {individual} has no valid mutations. "
            f"Returning original individual."
        )
        return lambda ind, config: ind

    mut_fn = random.choice(available_mutations)
    mut_fn(individual, config_space)
    return mut_fn



from typing import NamedTuple

DATA_TERMINAL = "data"


class Terminal(NamedTuple):
    """Specifies a specific value for a specific type or input.

    E.g. a value for a hyperparameter for an algorithm.

    It is important to note that you should use the hyperparameter's sklearn name as
    your output and identifier. If your name contains `__estimatorName`, you should
    remove it (e.g. by using string split). More information may be found in the
    documentation for the `get_hyperparameter_sklearn_name` function.

    Furthermore, the `config_space_name` is the name of the Config Space's
    hyperparameter. As a result, this is the name formed by the `__estimatorName` and
    the name of the hyperparameter.
    """

    value: object
    output: str
    identifier: str
    config_space_name: str = "Not Specified"

    def __str__(self) -> str:
        """str: e.g. "tol=0.5" """
        return f"{self.output}={format_hyperparameter_value(self.value)}"

    def __repr__(self) -> str:
        """str: e.g. "FastICA.tol=0.5".

        If the hyperparameter is shared across primitives, there is no prefix.
        """
        return f"{self.identifier}={format_hyperparameter_value(self.value)}"


def format_hyperparameter_value(value: object) -> str:
    if isinstance(value, str):
        return f"'{value}'"  # Quoted
    elif callable(value) and hasattr(value, "__name__"):
        return f"{value.__name__}"
    else:
        return str(value)



import functools
from typing import Any, Optional

import ConfigSpace
from ConfigSpace import Configuration
from smac import (
    Scenario,
    AlgorithmConfigurationFacade,
    RandomFacade,
    MultiFidelityFacade,
    HyperparameterOptimizationFacade,
    HyperbandFacade,
    BlackBoxFacade,
)
from smac.facade import AbstractFacade
from smac.initial_design import (
    RandomInitialDesign,
    DefaultInitialDesign,
    FactorialInitialDesign,
    SobolInitialDesign,
    LatinHypercubeInitialDesign,
)
from smac.main.config_selector import ConfigSelector

from gama.genetic_programming.components import Individual
from gama.genetic_programming.operations import _config_to_primitive_node
from gama.genetic_programming.longitudinal.operations_longitudinal import (
    _config_to_longitudinal_primitive_node,
)
from gama.genetic_programming.operator_set import OperatorSet

COMPONENTS_MAPPING = {
    "initial_design": {
        "RandomInitialDesign": RandomInitialDesign,
        "DefaultInitialDesign": DefaultInitialDesign,
        "FactorialInitialDesign": FactorialInitialDesign,
        "LatinHypercubeInitialDesign": LatinHypercubeInitialDesign,
        "SobolInitialDesign": SobolInitialDesign,
    },
    "facade": {
        "RandomFacade": RandomFacade,
        "MultiFidelityFacade": MultiFidelityFacade,
        "HyperparameterOptimizationFacade": HyperparameterOptimizationFacade,
        "HyperbandFacade": HyperbandFacade,
        "BlackBoxFacade": BlackBoxFacade,
        "AlgorithmConfigurationFacade": AlgorithmConfigurationFacade,
    },
}


def get_component(
    scenario: Scenario,
    component_type: str,
    component_mapping: dict,
    component_params: dict,
) -> Any:
    """Returns an instance of a component based on its type and parameters."""
    if component_type in component_mapping:
        comp_class = component_mapping[component_type]
    else:
        raise ValueError(
            f"BayesianOptimisation: Invalid component type "
            f"'{component_type}' passed to 'get_component'."
        )
    return comp_class(scenario=scenario, **component_params)


def get_smac(
    configSpace: ConfigSpace.ConfigurationSpace,
    scenario_params: Optional[dict] = None,
    initial_design_params: Optional[dict] = None,
    facade_params: Optional[dict] = None,
) -> AbstractFacade:
    if scenario_params:
        scenario = Scenario(configSpace, **scenario_params)
    else:
        scenario = Scenario(configSpace)

    if initial_design_params:
        if initial_design_params.get("type") is None:
            raise ValueError(
                "BayesianOptimisation: facade_params must contain a 'type' attribute."
            )
        initial_design_params_without_type = initial_design_params.copy()
        initial_design_params_without_type.pop("type")
        initial_design = get_component(
            scenario=scenario,
            component_type=initial_design_params["type"],
            component_mapping=COMPONENTS_MAPPING["initial_design"],
            component_params={**initial_design_params_without_type},
        )
    else:
        initial_design = RandomInitialDesign(scenario=scenario)

    if facade_params:
        if facade_params.get("type") is None:
            raise ValueError(
                "BayesianOptimisation: facade_params must contain a 'type' attribute."
            )
        facade_params_without_type = facade_params.copy()
        facade_params_without_type.pop("type")
        facade = get_component(
            scenario=scenario,
            component_type=facade_params["type"],
            component_mapping=COMPONENTS_MAPPING["facade"],
            component_params={
                **facade_params_without_type,
                "initial_design": initial_design,
                "target_function": dummy_smac_train,
            },
        )
    else:
        facade = AlgorithmConfigurationFacade(
            scenario=scenario,
            initial_design=initial_design,
            target_function=dummy_smac_train,
            logging_level=0,
            config_selector=ConfigSelector(scenario, retries=1000),
        )

    return facade


def dummy_smac_train(config: Configuration, seed: int = 0) -> None:
    _, _ = config, seed
    raise NotImplementedError(
        "BayesianOptimisation: dummy_smac_train should not have been called."
        "Current version does not support smac.Optimize() training procedure in "
        "favour to the Ask&Tell SMAC's interface. Operations.evaluate should be used "
        "instead through the GAMA multi-processing interface."
    )


def config_to_individual(
    config: Configuration,
    operations: OperatorSet,
) -> Individual:
    """Convert a SMAC configuration to a GAMA individual."""
    if config is None:
        raise ValueError(
            "BayesianOptimisation: config_to_individual received a None config."
        )
    if (config_space := operations.get_search_space()) is None:  # type: ignore
        raise ValueError(
            "BayesianOptimisation: Operations.get_search_space() returned None."
        )
    if (
        "preprocessors" in config_space.meta
        and config_space.meta["preprocessors"] in config.keys()
    ):
        max_length = 2
    else:
        max_length = 1
    return Individual(
        main_node=_config_to_primitive_node(
            config=config,
            config_meta=config_space.meta,
            conditions=config_space.get_conditions(),
            config_length=max_length,
        ),
        to_pipeline=operations._safe_compile or operations._compile,
    )


def config_to_longitudinal_individual(
    config: Configuration,
    operations: OperatorSet,
) -> Individual:
    """Convert a SMAC configuration to a GAMA individual."""
    if config is None:
        raise ValueError(
            "BayesianOptimisation: config_to_individual received a None config."
        )
    if (config_space := operations.get_search_space()) is None:  # type: ignore
        raise ValueError(
            "BayesianOptimisation: Operations.get_search_space() returned None."
        )
    if (
        "preprocessors" in config_space.meta
        and config_space.meta["preprocessors"] in config.keys()
    ):
        max_length = 2
    else:
        max_length = 1
    return Individual(
        main_node=_config_to_longitudinal_primitive_node(
            config=config,
            config_meta=config_space.meta,
            conditions=config_space.get_conditions(),
            config_length=max_length,
        ),
        to_pipeline=operations._safe_compile or operations._compile,
    )


def validate_info_config(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if not hasattr(result, "config") or result.config is None:
            raise ValueError(
                f"BayesianOptimisation: Function "
                f"'{func.__name__}' returned info with None config."
            )
        return result

    return wrapper


def validate_future_valid(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        future = args[0]
        if future is None or not hasattr(future, "result") or future.result is None:
            raise ValueError(
                f"BayesianOptimisation: Function '{func.__name__}' "
                f"received a Future object or its result is not valid."
            )

        result = future.result
        if not hasattr(result, "individual") or result.individual is None:
            raise ValueError(
                "BayesianOptimisation: Future object's result is missing 'individual' "
                "attribute."
            )

        if (
            not hasattr(result.individual, "fitness")
            or result.individual.fitness is None
        ):
            raise ValueError(
                "BayesianOptimisation: Future object's result's individual is missing "
                "'fitness' attribute."
            )

        required_attrs = ["values", "wallclock_time", "start_time", "process_time"]

        for attr in required_attrs:
            if not hasattr(result.individual.fitness, attr):
                raise ValueError(
                    f"BayesianOptimisation: Future object's result's individual's "
                    f"fitness is missing '{attr}' attribute."
                )

        if hasattr(result.individual.fitness, "start_time") and not callable(
            getattr(result.individual.fitness.start_time, "timestamp", None)
        ):
            raise ValueError(
                "BayesianOptimisation: start_time object does not have a timestamp "
                "method."
            )

        return func(*args, **kwargs)

    return wrapper



import ConfigSpace as cs

from .regression_task import RegressorConfig, PreprocessorConfig

# Regressors & Preprocessors 🚀

# This script is your ticket to configuring a ConfigSpace object, teeming with
# regressors and preprocessors. We are diving in with the RegressorConfig and
# PreprocessorConfig classes to fill the configuration space with a slew of
# hyperparameters and options.

# Customise Your Space 🔧

# Want just classifiers? No biggie! Just comment out or remove the PreprocessorConfig
# setup. Voila! You're left with a sleek, regressor-only configuration space.

# Want to add more regressors or preprocessors? Easy! Just add them to the
# RegressorConfig or PreprocessorConfig classes, respectively. You can even
# add your own custom regressors or preprocessors. Just make sure they are
# compatible with scikit-learn's API.

# Meta-Parameters 📝

# The meta-parameters are the "estimators" and "preprocessors" keys in the
# configuration space. These are used to identify the regressors and preprocessors
# by the internal system. They are not hyperparameters, and should not be
# changed, except by advanced users. If you do change them, make sure to change
# the corresponding values in the current configuration space, i.e. in RegressorConfig
# and PreprocessorConfig.

# 👩‍💻👨‍💻 Happy configuring, and may your machine learning models shine!

config_space = cs.ConfigurationSpace(
    meta={
        # "gama_system_name": "current_configuration_name",
        "estimators": "regressors",
        "preprocessors": "preprocessors",
    }
)

regressor_config = RegressorConfig(config_space)
regressor_config.setup_regressors()

preprocessor_config = PreprocessorConfig(config_space)
preprocessor_config.setup_preprocessors()



import ConfigSpace as cs

from gama.configuration.configuration_task_test import (
    ClassifierConfigTest,
    PreprocessorConfigTest,
)

# A configuration with limited operators for unit tests 🧪

config_space = cs.ConfigurationSpace(
    meta={
        # "gama_system_name": "current_configuration_name",
        "estimators": "classifiers",
        "preprocessors": "preprocessors",
    }
)

classifier_config = ClassifierConfigTest(config_space)
classifier_config.setup_classifiers()

preprocessor_config = PreprocessorConfigTest(config_space)
preprocessor_config.setup_preprocessors()



import copy
from typing import Tuple, List, Set, Optional

from sklearn.base import TransformerMixin

from gama.GamaPipeline import GamaPipelineTypeUnion, GamaPipelineType
from gama.genetic_programming.components import Individual


def transformers_to_str(transformers: List[TransformerMixin]) -> List[str]:
    """Format a transformer for code export, removes any mapping."""
    copies = list(map(copy.copy, transformers))
    for transformer in copies:
        if hasattr(transformer, "mapping"):
            transformer.mapping = None
    return list(map(str, copies))


def format_import(o: object) -> str:
    """Creates the import statement for `o`'s class."""
    if o.__module__.split(".")[-1].startswith("_"):
        module = ".".join(o.__module__.split(".")[:-1])
    else:
        module = o.__module__
    return f"from {module} import {o.__class__.__name__}"


def format_pipeline(steps: List[Tuple[str, str]], name: str = "pipeline") -> str:
    steps_str = ",\n".join([f"('{name}', {step})" for name, step in steps])
    return f"{name} = Pipeline([{steps_str}])\n"


def imports_and_steps_for_individual(
    individual: Individual,
    gama_pipeline_type: Optional[GamaPipelineTypeUnion] = GamaPipelineType.ScikitLearn,
) -> Tuple[Set[str], List[Tuple[str, str]]]:
    """Determine required imports and steps for the individual's pipeline.

    Returns two lists:
     - one with import statements
     - one with tuples with pipeline step names and step values

    E.g. (["from sklearn.naive_bayes import GaussianNB"], [('0', 'GaussianNB()')])
    """
    imports = [
        "from numpy import nan",
    ]
    if gama_pipeline_type:
        imports.append(GamaPipelineType(gama_pipeline_type).required_import)
    imports += [format_import(step) for name, step in individual.pipeline.steps]

    steps = []
    for i, primitive_node in reversed(list(enumerate(individual.primitives))):
        steps.append((str(i), primitive_node.str_nonrecursive))
        imports.extend(
            f"from {terminal.value.__module__} import {terminal.value.__name__}"
            for terminal in primitive_node._terminals
            if callable(terminal.value) and hasattr(terminal.value, "__name__")
        )
    return set(imports), steps


def individual_to_python(
    individual: Individual,
    prepend_steps: Optional[List[Tuple[str, TransformerMixin]]] = None,
    gama_pipeline_type: Optional[GamaPipelineTypeUnion] = GamaPipelineType.ScikitLearn,
) -> str:
    """Generate code for the machine learning pipeline represented by `individual`."""
    imports, steps = imports_and_steps_for_individual(individual, gama_pipeline_type)
    if prepend_steps is not None:
        steps = prepend_steps + steps
        imports = imports.union({format_import(step) for _, step in prepend_steps})
    steps_str = ",\n".join([f"('{name}', {step})" for name, step in steps])
    pipeline = f"Pipeline([{steps_str}])"
    return "\n".join(sorted(imports)) + "\n\n" + "pipeline = " + pipeline + "\n"



import logging
from typing import Callable, Dict, List, Optional, Tuple, Any

import ConfigSpace

from gama.genetic_programming.components.primitive_node import PrimitiveNode


from gama.utilities.evaluation_library import Evaluation

from .components import Individual
from ..GamaPipeline import GamaPipelineTypeUnion

log = logging.getLogger(__name__)


class OperatorSet:
    """Provides a thin layer for ea operators for logging, callbacks and safety."""

    def __init__(
        self,
        mutate: Callable[[Individual], None],
        mate: Callable[[Individual, Individual], Tuple[Individual, Individual]],
        create_from_population: Callable[[Any], List[Individual]],
        create_new: Callable[[], PrimitiveNode],
        compile_: Callable[[Individual], GamaPipelineTypeUnion],
        eliminate: Callable[[List[Individual], int], List[Individual]],
        evaluate_callback: Callable[[Evaluation], None],
        max_retry: int = 50,
        completed_evaluations: Optional[Dict[str, Evaluation]] = None,
        is_evaluated: Optional[Callable[[Individual], bool]] = None,
        get_search_space: Optional[Callable[[], ConfigSpace.ConfigurationSpace]] = None,
    ):
        self._mutate = mutate
        self._mate = mate
        self._create_from_population = create_from_population
        self._create_new = create_new
        self._compile = compile_
        self._safe_compile: Optional[
            Callable[[Individual], GamaPipelineTypeUnion]
        ] = None
        self._eliminate = eliminate
        self._max_retry = max_retry
        self._evaluate = None
        self._evaluate_callback = evaluate_callback
        self.evaluate: Optional[Callable[..., Evaluation]] = None
        self.is_evaluated = is_evaluated
        self.get_search_space = get_search_space

        self._completed_evaluations = completed_evaluations

    def wait_next(self, async_evaluator):
        """Wrapper for wait_next() to forward evaluation and log exceptions."""
        future = async_evaluator.wait_next()
        if future.result and self._evaluate_callback:
            self._evaluate_callback(future.result)
        elif future.exception:
            log.warning(f"Error raised during evaluation: {str(future.exception)}.")
        return future

    def try_until_new(self, operator, *args, **kwargs):
        """Keep executing `operator` until a new individual is created."""
        for _ in range(self._max_retry):
            individual = operator(*args, **kwargs)
            if str(individual.main_node) not in self._completed_evaluations:
                return individual
        log.debug(f"50 iterations of {operator.__name__} did not yield new ind.")
        # For progress on solving this, see #11
        return individual

    def mate(self, ind1: Individual, ind2: Individual, *args, **kwargs) -> Individual:
        def mate_with_log():
            new_individual1, new_individual2 = ind1.copy_as_new(), ind2.copy_as_new()
            self._mate(new_individual1, new_individual2, *args, **kwargs)
            new_individual1.meta = dict(parents=[ind1._id, ind2._id], origin="cx")
            return new_individual1

        return self.try_until_new(mate_with_log)

    def mutate(self, ind: Individual, *args, **kwargs) -> Individual:
        def mutate_with_log():
            new_individual = ind.copy_as_new()
            mutator = self._mutate(new_individual, *args, **kwargs)
            new_individual.meta = dict(parents=[ind._id], origin=mutator.__name__)
            return new_individual

        return self.try_until_new(mutate_with_log)

    def individual(self, *args, **kwargs) -> Individual:
        expression = self._create_new(*args, **kwargs)
        compile_ = self._safe_compile or self._compile
        ind = Individual(expression, to_pipeline=compile_)
        ind.meta["origin"] = "new"
        return ind

    def create(self, *args, **kwargs) -> List[Individual]:
        return self._create_from_population(self, *args, **kwargs)

    def eliminate(self, *args, **kwargs):
        return self._eliminate(*args, **kwargs)



import copy
import random
from typing import List, Optional, Any, Union, Callable, Tuple

import ConfigSpace as cs

from gama.genetic_programming.components import (
    Primitive,
    PrimitiveNode,
    DATA_TERMINAL,
    Terminal,
)
from gama.utilities.config_space import (
    get_internal_output_types,
    get_hyperparameter_sklearn_name,
    get_estimator_by_name,
)


def _sample_and_ignore_dummy_techniques(
    config_space: cs.ConfigurationSpace,
    config: cs.Configuration,
    technique_name_to_ignore: str = "Dummy_To_Ignore",
) -> cs.Configuration:
    """ConfigSpace do not allow adding forbidden
    clauses on the default choice of a hyperparameter.

    Dummy_To_Ignore becomes default and is ignored / found
    an alternative when picked by sample_configuration.
    Nonetheless, for better maintainability, "Dummy_To_Ignore" can be changed,
    for example in the future it could be in the meta of the config so that
    it can be changed in one place by the user, i.e in /configuratins.
    """
    is_preprocessor_to_ignore = (
        "preprocessors" in config_space.meta
        and config[config_space.meta["preprocessors"]] == technique_name_to_ignore
    )
    is_estimator_to_ignore = (
        "estimators" in config_space.meta
        and config[config_space.meta["estimators"]] == technique_name_to_ignore
    )
    if is_estimator_to_ignore or is_preprocessor_to_ignore:
        while is_estimator_to_ignore or is_preprocessor_to_ignore:
            temp_config = config_space.sample_configuration()
            is_preprocessor_to_ignore = (
                "preprocessors" in config_space.meta
                and temp_config[config_space.meta["preprocessors"]]
                == technique_name_to_ignore
            )
            is_estimator_to_ignore = (
                "estimators" in config_space.meta
                and temp_config[config_space.meta["estimators"]]
                == technique_name_to_ignore
            )
            if not is_estimator_to_ignore and not is_preprocessor_to_ignore:
                return temp_config
    return config


def random_primitive_node(
    output_type: str,
    config_space: cs.ConfigurationSpace,
    exclude: Optional[Primitive] = None,
) -> PrimitiveNode:
    """Create a PrimitiveNode with specified output_type and random terminals."""
    if output_type not in get_internal_output_types():
        raise ValueError(f"Output type {output_type} not supported")

    if exclude is not None:
        temp_config_space = copy.deepcopy(config_space)
        if output_type not in temp_config_space.meta:
            raise ValueError(f"Output type {output_type} not in config_space meta.")
        temp_config_space.add_forbidden_clause(
            cs.ForbiddenEqualsClause(
                temp_config_space.get_hyperparameter(
                    temp_config_space.meta[output_type]
                ),
                exclude.__str__(),
            )
        )
        config = _sample_and_ignore_dummy_techniques(
            temp_config_space, temp_config_space.sample_configuration()
        )
    else:
        config = _sample_and_ignore_dummy_techniques(
            config_space, config_space.sample_configuration()
        )

    if output_type in [DATA_TERMINAL, "preprocessors"]:
        (
            preprocessor_primitive,
            preprocessor_terminals,
        ) = _config_component_to_primitive_node(
            "preprocessors",
            config,
            config_space.meta,
            config_space.get_conditions(),
            ["estimators", "preprocessors"],
        )
        return PrimitiveNode(
            preprocessor_primitive,
            data_node=DATA_TERMINAL,
            terminals=preprocessor_terminals,
        )
    estimator_primitive, estimator_terminals = _config_component_to_primitive_node(
        "estimators",
        config,
        config_space.meta,
        config_space.get_conditions(),
        ["estimators", "preprocessors"],
    )
    return PrimitiveNode(
        primitive=estimator_primitive,
        data_node=DATA_TERMINAL,
        terminals=estimator_terminals,
    )


def create_random_expression(
    config_space: cs.ConfigurationSpace,
    min_length: int = 1,
    max_length: int = 3,
) -> PrimitiveNode:
    """Create at least min_length and at most max_length chained PrimitiveNodes."""
    individual_length = random.randint(min_length, max_length)
    config = _sample_and_ignore_dummy_techniques(
        config_space, config_space.sample_configuration()
    )
    return _config_to_primitive_node(
        config=config,
        config_meta=config_space.meta,
        conditions=config_space.get_conditions(),
        config_length=individual_length,
    )


def _config_to_primitive_node(
    config: cs.Configuration,
    config_meta: dict,
    conditions: List[Any],
    config_length: Optional[int] = None,
) -> PrimitiveNode:
    """Create a PrimitiveNode from a configuration. If config_length is specified, the
    PrimitiveNode will have at most config_length PrimitiveNodes."""
    if isinstance(config_length, int) and config_length <= 1:
        estimator_primitive, estimator_terminals = _config_component_to_primitive_node(
            "estimators",
            config,
            config_meta,
            conditions,
            ["estimators", "preprocessors"],
        )
        return PrimitiveNode(
            estimator_primitive, data_node=DATA_TERMINAL, terminals=estimator_terminals
        )
    estimator_primitive, estimator_terminals = _config_component_to_primitive_node(
        "estimators", config, config_meta, conditions, ["estimators", "preprocessors"]
    )

    (
        preprocessor_primitive,
        preprocessor_terminals,
    ) = _config_component_to_primitive_node(
        "preprocessors",
        config,
        config_meta,
        conditions,
        ["estimators", "preprocessors"],
    )

    # Create a PrimitiveNode for the preprocessor
    preprocessor_node = PrimitiveNode(
        preprocessor_primitive,
        data_node=DATA_TERMINAL,
        terminals=preprocessor_terminals,
    )

    # Create a PrimitiveNode for the classifier and chain it to the preprocessor
    return PrimitiveNode(
        estimator_primitive, data_node=preprocessor_node, terminals=estimator_terminals
    )


def _config_component_to_primitive_node(
    component_type: str,
    config: cs.Configuration,
    config_meta: dict,
    conditions: List[Any],
    exclude_keys: List[str],
    output_type: Optional[str] = None,
    retrieve_estimator: Callable = get_estimator_by_name,
) -> Tuple[Primitive, List[Terminal]]:
    """
    Create a PrimitiveNode from a configuration of type ConfigSpace
    (estimator or preprocessor).

    This function generalizes the creation of a PrimitiveNode for either an estimator or
    a preprocessor, based on the component type specified. It creates a Primitive for
    the selected component, determines the valid hyperparameters for the component
    based on the conditions, and then creates a Terminal for each valid hyperparameter.

    Parameters
    ----------
    component_type : str
        The type of component (e.g., 'estimators' or 'preprocessors').
    config : cs.Configuration
        A configuration of type ConfigSpace.
    config_meta : dict
        The meta information of the ConfigSpace.
    conditions : List[Any]
        The conditions of the ConfigSpace.
    exclude_keys : List[str]
        Keys to exclude when considering hyperparameters.
    output_type : str, optional
        The output type of the PrimitiveNode, by default None.
    """
    if (
        component_type not in config_meta
        or config_meta[component_type] not in config.get_dictionary()
    ):
        raise ValueError(
            f"Configuration {config} does not contain a `{component_type}` ConfigSpace"
            "Hyperparameter. Cannot construct PrimitiveNode."
        )

    if output_type is None:
        output_type = component_type

    # Create a Primitive for the selected component
    component_primitive = Primitive(
        identifier=retrieve_estimator(config[config_meta[component_type]]),
        output=output_type,
        input=tuple(
            get_hyperparameter_sklearn_name(hp)
            for hp in config
            if hp not in [config_meta.get(key) for key in exclude_keys]
        ),
    )

    # Determine valid hyperparameters for the component based on conditions
    component_valid_hyperparameters = [
        name
        for condition in conditions
        if (
            name := extract_valid_hyperparameters(
                condition, config, config_meta, component_type
            )
        )
        is not None
    ]

    # Create a Terminal for each valid hyperparameter for the component
    component_terminals = [
        Terminal(
            identifier=get_hyperparameter_sklearn_name(param),
            value=value,
            output=get_hyperparameter_sklearn_name(param),
            config_space_name=param,
        )
        for param, value in config.items()
        if param in component_valid_hyperparameters
        and param not in [config_meta.get(key) for key in exclude_keys]
    ]

    return component_primitive, component_terminals


def extract_valid_hyperparameters(
    cond: cs.conditions, config: cs.Configuration, config_meta: dict, meta_key: str
) -> Union[str, None]:
    """Extract valid hyperparameters from a condition.

    For each supported ConfigSpace condition type, extract the valid hyperparameters
    from the condition. The valid hyperparameters are the hyperparameters that are
    valid for the given condition and configuration based on the meta_key.

    Supported ConfigSpace condition types:
    - EqualsCondition
    - AndConjunction

    Readers are encouraged to add support for more ConfigSpace condition types if
    needed. Open an issue or pull request on the GAMA GitHub repository.

    Parameters
    ----------
    cond : cs.conditions
        A condition of type ConfigSpace.
    config : cs.Configuration
        A configuration of type ConfigSpace.
    config_meta : dict
        The meta information of the ConfigSpace.
    meta_key : str
        The meta key of the ConfigSpace.
    """
    if meta_key not in config_meta:
        raise ValueError(f"Meta key {meta_key} not in config_meta")
    if type(cond) not in [cs.conditions.EqualsCondition, cs.conditions.AndConjunction]:
        raise ValueError(
            f"Condition type {type(cond)} not supported. Refer to "
            f"docstring for supported condition types."
        )
    if isinstance(cond, cs.conditions.EqualsCondition):
        if (
            cond.parent.name == config_meta[meta_key]
            and cond.value == config[config_meta[meta_key]]
        ):
            return cond.child.name
    elif isinstance(cond, cs.conditions.AndConjunction):
        for component in cond.components:
            if (
                component.parent.name == config_meta[meta_key]
                and component.value == config[config_meta[meta_key]]
            ):
                return component.child.name
    return None



import logging
import os
import time
from datetime import datetime
from typing import Callable, Tuple, Optional, Sequence

import numpy as np
import stopit
from sklearn.base import TransformerMixin, is_classifier
from sklearn.model_selection import (
    ShuffleSplit,
    cross_validate,
    check_cv,
    StratifiedShuffleSplit,
)

from gama.GamaPipeline import GamaPipelineTypeUnion, GamaPipeline, GamaPipelineType
from gama.genetic_programming.components import Individual, PrimitiveNode, Fitness
from gama.utilities.evaluation_library import Evaluation
from gama.utilities.generic.stopwatch import Stopwatch
from gama.utilities.metrics import Metric

log = logging.getLogger(__name__)


def primitive_node_to_sklearn(primitive_node: PrimitiveNode) -> object:
    hyperparameters = {
        terminal.output: terminal.value for terminal in primitive_node._terminals
    }
    return primitive_node._primitive.identifier(**hyperparameters)


def compile_individual(
    individual: Individual,
    _parameter_checks=None,
    preprocessing_steps: Optional[Sequence[Tuple[str, TransformerMixin]]] = None,
    gama_pipeline_type: Optional[GamaPipelineTypeUnion] = GamaPipelineType.ScikitLearn,
    *args,
    **kwargs,
) -> GamaPipelineTypeUnion:
    steps = [
        (str(i), primitive_node_to_sklearn(primitive))
        for i, primitive in enumerate(individual.primitives)
    ]
    if preprocessing_steps:
        steps += list(reversed(preprocessing_steps))
    if gama_pipeline_type:
        return GamaPipeline(
            list(reversed(steps)),
            pipeline_type=gama_pipeline_type,
            *args,
            **kwargs,  # type: ignore
        )
    return GamaPipeline(list(reversed(steps)))


def object_is_valid_pipeline(o: object) -> bool:
    """Determines if object behaves like a scikit-learn pipeline."""
    return (
        o is not None
        and hasattr(o, "fit")
        and hasattr(o, "predict")
        and hasattr(o, "steps")
    )


def evaluate_pipeline(
    pipeline,
    x,
    y_train,
    timeout: float,
    metrics: Tuple[Metric],
    cv=5,
    subsample=None,
) -> Tuple:
    """Score `pipeline` with k-fold CV according to `metrics` on (a subsample of) X, y

    Returns
    -------
    Tuple:
        prediction: np.ndarray if successful, None if not
        scores: tuple with one float per metric, each value is -inf on fail.
        estimators: list of fitted pipelines if successful, None if not
        error: None if successful, otherwise an Exception
    """
    if not object_is_valid_pipeline(pipeline):
        raise TypeError("Pipeline must not be None and requires fit, predict, steps.")
    if timeout <= 0:
        raise ValueError(f"`timeout` must be greater than 0, is {timeout}.")

    prediction, estimators = None, None
    # default score for e.g. timeout or failure
    scores = tuple([float("-inf")] * len(metrics))
    is_classification = is_classifier(pipeline)

    with stopit.ThreadingTimeout(timeout) as c_mgr:
        try:
            # When splits are generated (i.e., cv is an int), they are deterministic
            splitter = check_cv(cv, y_train, classifier=is_classification)

            require_subsample = (
                isinstance(subsample, int) and subsample < len(y_train)
            ) or (isinstance(subsample, float) and subsample < 1.0)

            if require_subsample:
                # We subsample the training sets, but not the test sets.
                # This allows for performance comparisons across subsample levels.
                new_splits = []
                for train, test in splitter.split(x, y_train):
                    if is_classification:
                        sampler = StratifiedShuffleSplit(
                            n_splits=1, train_size=subsample, random_state=0
                        )
                    else:
                        sampler = ShuffleSplit(
                            n_splits=1, train_size=subsample, random_state=0
                        )
                    full_train_x, full_train_y = x.iloc[train, :], y_train[train]
                    subsample_idx, _ = next(sampler.split(full_train_x, full_train_y))
                    new_splits.append((subsample_idx, test))
                splitter = new_splits

            result = cross_validate(
                pipeline,
                x,
                y_train,
                cv=splitter,
                return_estimator=True,
                scoring=dict([(m.name, m) for m in metrics]),
                error_score="raise",
            )
            scores = tuple(np.mean(result[f"test_{m.name}"]) for m in metrics)
            estimators = result["estimator"]

            splitter = (
                splitter if isinstance(splitter, list) else splitter.split(x, y_train)
            )
            for estimator, (_, test) in zip(estimators, splitter):
                if any(m.requires_probabilities for m in metrics):
                    fold_pred = estimator.predict_proba(x.iloc[test, :])
                else:
                    fold_pred = estimator.predict(x.iloc[test, :])

                if prediction is None:
                    if fold_pred.ndim == 2:
                        prediction = np.empty(shape=(len(y_train), fold_pred.shape[1]))
                    else:
                        prediction = np.empty(shape=(len(y_train),))
                prediction[test] = fold_pred

        except stopit.TimeoutException:
            # This exception is handled by the ThreadingTimeout context manager.
            raise
        except KeyboardInterrupt:
            raise
        except Exception as e:
            return prediction, scores, estimators, e

    if c_mgr.state == c_mgr.INTERRUPTED:
        # A TimeoutException was raised, but not by the context manager.
        # This indicates that the outer context manager (the ea) timed out.
        raise stopit.utils.TimeoutException()

    return (
        (prediction, tuple(scores), estimators, None)
        if c_mgr
        else (prediction, scores, estimators, stopit.TimeoutException())
    )


def evaluate_individual(
    individual: Individual,
    evaluate_pipeline: Callable,
    timeout: float = 1e6,
    deadline: Optional[float] = None,
    add_length_to_score: bool = True,
    **kwargs,
) -> Evaluation:
    """Evaluate the pipeline specified by individual, and record

    Parameters
    ----------
    individual: Individual
        Blueprint for the pipeline to evaluate.
    evaluate_pipeline: Callable
        Function which takes the pipeline and produces validation predictions,
        scores, estimators and errors.
    timeout: float (default=1e6)
        Maximum time in seconds that the evaluation is allowed to take.
        Don't depend on high accuracy.
        A shorter timeout is imposed if `deadline` is in less than `timeout` seconds.
    deadline: float, optional
        A time in seconds since epoch.
        Cut off evaluation at `deadline` even if `timeout` seconds have not yet elapsed.
    add_length_to_score: bool (default=True)
        Add the length of the individual to the score result of the evaluation.
    **kwargs: Dict, optional (default=None)
        Passed to `evaluate_pipeline` function.

    Returns
    -------
    Evaluation

    """
    result = Evaluation(individual, pid=os.getpid())
    result.start_time = datetime.now()

    if deadline is not None:
        time_to_deadline = deadline - time.time()
        timeout = min(timeout, time_to_deadline)

    with Stopwatch() as wall_time, Stopwatch(time.process_time) as process_time:
        evaluation = evaluate_pipeline(individual.pipeline, timeout=timeout, **kwargs)
        result._predictions, result.score, result._estimators, error = evaluation
        if error is not None:
            result.error = f"{type(error)} {str(error)}"
    result.duration = wall_time.elapsed_time

    if add_length_to_score:
        result.score = result.score + (-len(individual.primitives),)
    individual.fitness = Fitness(
        result.score,
        result.start_time,
        wall_time.elapsed_time,
        process_time.elapsed_time,
    )

    return result



import copy
import random
from typing import List, Optional, Any

import ConfigSpace as cs

from gama.genetic_programming.components import (
    Primitive,
    PrimitiveNode,
    DATA_TERMINAL,
)
from gama.genetic_programming.operations import _config_component_to_primitive_node
from gama.utilities.config_space import (
    get_longitudinal_estimator_by_name,
    get_internal_longitudinal_output_types,
)


def _sample_and_ignore_dummy_techniques(
    config_space: cs.ConfigurationSpace,
    config: cs.Configuration,
    technique_name_to_ignore: str = "Dummy_To_Ignore",
) -> cs.Configuration:
    """ConfigSpace do not allow adding forbidden
    clauses on the default choice of a hyperparameter.

    Dummy_To_Ignore becomes default and is ignored / found
    an alternative when picked by sample_configuration.
    Nonetheless, for better maintainability, "Dummy_To_Ignore" can be changed,
    for example in the future it could be in the meta of the config so that
    it can be changed in one place by the user, i.e in /configuratins.
    """
    is_data_preparation_to_ignore = (
        "data_preparation" in config_space.meta
        and config[config_space.meta["data_preparation"]] == technique_name_to_ignore
    )
    is_preprocessor_to_ignore = (
        "preprocessors" in config_space.meta
        and config[config_space.meta["preprocessors"]] == technique_name_to_ignore
    )
    is_estimator_to_ignore = (
        "estimators" in config_space.meta
        and config[config_space.meta["estimators"]] == technique_name_to_ignore
    )

    if (
        is_data_preparation_to_ignore
        or is_estimator_to_ignore
        or is_preprocessor_to_ignore
    ):
        while (
            is_data_preparation_to_ignore
            or is_estimator_to_ignore
            or is_preprocessor_to_ignore
        ):
            temp_config = config_space.sample_configuration()
            is_data_preparation_to_ignore = (
                "data_preparation" in config_space.meta
                and temp_config[config_space.meta["data_preparation"]]
                == technique_name_to_ignore
            )
            is_preprocessor_to_ignore = (
                "preprocessors" in config_space.meta
                and temp_config[config_space.meta["preprocessors"]]
                == technique_name_to_ignore
            )
            is_estimator_to_ignore = (
                "estimators" in config_space.meta
                and temp_config[config_space.meta["estimators"]]
                == technique_name_to_ignore
            )
            if (
                not is_data_preparation_to_ignore
                and not is_estimator_to_ignore
                and not is_preprocessor_to_ignore
            ):
                return temp_config
    return config


def random_longitudinal_primitive_node(
    output_type: str,
    config_space: cs.ConfigurationSpace,
    exclude: Optional[Primitive] = None,
    exclude_options_list: Optional[List[str]] = None,
) -> PrimitiveNode:
    """Create a PrimitiveNode with specified output_type and random terminals.

    copy from gama/genetic_programming/operations.py with some modification, which are:
    1. We use longitudinal output types check instead of normal output types check
    2. We now support exclude_options_list, which is a list of options that we want to
    exclude from the search space but cannot be represented within ConfigSpace given
    other conditions. We then repetitively sample from the search space until we get
    a valid option, which is always possible to find.
    3. _config_component_to_primitive_node uses a different exclude_keys list, which
    is ["data_preparation", "estimators", "preprocessors"] instead of
    ["estimators", "preprocessors"]
    """
    if output_type not in get_internal_longitudinal_output_types():
        raise ValueError(f"Output type {output_type} not supported")
    excludes = exclude_options_list.copy() if exclude_options_list else None

    if exclude is not None or excludes is not None:
        temp_config_space = copy.deepcopy(config_space)
        if output_type not in temp_config_space.meta:
            raise ValueError(f"Output type {output_type} not in config_space meta.")

        exclude_options_path = True
        if excludes:
            if exclude is not None:
                excludes.append(exclude.__str__())
            try:
                temp_config_space.add_forbidden_clause(
                    cs.ForbiddenInClause(
                        temp_config_space.get_hyperparameter(
                            temp_config_space.meta[output_type]
                        ),
                        excludes,
                    )
                )
                exclude_options_path = False
            except ValueError:
                exclude_options_path = True
        elif exclude:
            temp_config_space.add_forbidden_clause(
                cs.ForbiddenEqualsClause(
                    temp_config_space.get_hyperparameter(
                        temp_config_space.meta[output_type]
                    ),
                    exclude.__str__(),
                )
            )

        config = _sample_and_ignore_dummy_techniques(
            temp_config_space, temp_config_space.sample_configuration()
        )
        if exclude_options_path and excludes:
            while config[config_space.meta[output_type]] in excludes:
                config = _sample_and_ignore_dummy_techniques(
                    temp_config_space, temp_config_space.sample_configuration()
                )

    else:
        config = _sample_and_ignore_dummy_techniques(
            config_space, config_space.sample_configuration()
        )

    (
        data_preparation_primitive,
        data_preparation_terminals,
    ) = _config_component_to_primitive_node(
        "data_preparation",
        config,
        config_space.meta,
        config_space.get_conditions(),
        ["data_preparation", "estimators", "preprocessors"],
        retrieve_estimator=get_longitudinal_estimator_by_name,
    )
    if output_type == "data_preparation":
        return PrimitiveNode(
            data_preparation_primitive,
            data_node=DATA_TERMINAL,
            terminals=data_preparation_terminals,
        )
    if output_type in [DATA_TERMINAL, "preprocessors"]:
        (
            preprocessor_primitive,
            preprocessor_terminals,
        ) = _config_component_to_primitive_node(
            "preprocessors",
            config,
            config_space.meta,
            config_space.get_conditions(),
            ["data_preparation", "estimators", "preprocessors"],
            retrieve_estimator=get_longitudinal_estimator_by_name,
        )
        data_preparation_node = PrimitiveNode(
            data_preparation_primitive,
            data_node=DATA_TERMINAL,
            terminals=data_preparation_terminals,
        )
        return PrimitiveNode(
            preprocessor_primitive,
            data_node=data_preparation_node,
            terminals=preprocessor_terminals,
        )
    estimator_primitive, estimator_terminals = _config_component_to_primitive_node(
        "estimators",
        config,
        config_space.meta,
        config_space.get_conditions(),
        ["data_preparation", "estimators", "preprocessors"],
        retrieve_estimator=get_longitudinal_estimator_by_name,
    )
    data_preparation_node = PrimitiveNode(
        data_preparation_primitive,
        data_node=DATA_TERMINAL,
        terminals=data_preparation_terminals,
    )
    return PrimitiveNode(
        estimator_primitive,
        data_node=data_preparation_node,
        terminals=estimator_terminals,
    )


def create_longitudinal_random_expression(
    config_space: cs.ConfigurationSpace,
    min_length: int = 2,
    max_length: int = 3,
) -> PrimitiveNode:
    """Create at least min_length and at most max_length chained PrimitiveNodes.

    copy from gama/genetic_programming/operations.py with some modification, which are:
    1. Min length is 2 instead of 1, because we need at least one data preparation
    and one estimator.
    """
    individual_length = random.randint(min_length, max_length)
    config = _sample_and_ignore_dummy_techniques(
        config_space, config_space.sample_configuration()
    )
    return _config_to_longitudinal_primitive_node(
        config=config,
        config_meta=config_space.meta,
        conditions=config_space.get_conditions(),
        config_length=individual_length,
    )


def _config_to_longitudinal_primitive_node(
    config: cs.Configuration,
    config_meta: dict,
    conditions: List[Any],
    config_length: Optional[int] = None,
) -> PrimitiveNode:
    """Create a PrimitiveNode from a configuration. If config_length is specified, the
    PrimitiveNode will have at most config_length PrimitiveNodes.

    copy from gama/genetic_programming/operations.py with some modification, which are:
    1. We prepare the data_preparation and estimator first.
    2. We use _config_component_to_primitive_node with a different exclude_keys list,
    which is ["data_preparation", "estimators", "preprocessors"] instead of
    ["estimators", "preprocessors"]
    3. If the preprocessor option is None, we don't create a preprocessor node.
    4. If config_length is 2, we don't create a preprocessor node.
    """
    (
        data_preparation_primitive,
        data_preparation_terminals,
    ) = _config_component_to_primitive_node(
        "data_preparation",
        config,
        config_meta,
        conditions,
        ["data_preparation", "estimators", "preprocessors"],
        retrieve_estimator=get_longitudinal_estimator_by_name,
    )
    estimator_primitive, estimator_terminals = _config_component_to_primitive_node(
        "estimators",
        config,
        config_meta,
        conditions,
        ["data_preparation", "estimators", "preprocessors"],
        retrieve_estimator=get_longitudinal_estimator_by_name,
    )
    data_preparation_node = PrimitiveNode(
        data_preparation_primitive,
        data_node=DATA_TERMINAL,
        terminals=data_preparation_terminals,
    )
    if isinstance(config_length, int) and config_length <= 2:
        # Create a PrimitiveNode for the data_preparation
        return PrimitiveNode(
            estimator_primitive,
            data_node=data_preparation_node,
            terminals=estimator_terminals,
        )
    preprocessor_node = None
    if config[config_meta["preprocessors"]] != "None":
        (
            preprocessor_primitive,
            preprocessor_terminals,
        ) = _config_component_to_primitive_node(
            "preprocessors",
            config,
            config_meta,
            conditions,
            ["data_preparation", "estimators", "preprocessors"],
            retrieve_estimator=get_longitudinal_estimator_by_name,
        )
        # Create a PrimitiveNode for the preprocessor and chain it to the
        # data_preparation
        preprocessor_node = PrimitiveNode(
            preprocessor_primitive,
            data_node=data_preparation_node,
            terminals=preprocessor_terminals,
        )

    # Create a PrimitiveNode for the classifier and chain it to the preprocessor
    return PrimitiveNode(
        estimator_primitive,
        data_node=preprocessor_node or data_preparation_node,
        terminals=estimator_terminals,
    )



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class ClassifierConfigTest:
    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "estimators" not in config_space.meta:
            raise ValueError("Expected 'estimators' key in meta of config_space")
        self.config_space = config_space
        self.classifiers_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "BernoulliNB": self.setup_bernoulliNB,
            "MultinomialNB": self.setup_multinomialNB,
            "GaussianNB": self.setup_gaussianNB,
            "DecisionTreeClassifier": self.setup_decision_tree,
            "ExtraTreesClassifier": self.setup_extra_trees,
            "RandomForestClassifier": self.setup_random_forest,
            "GradientBoostingClassifier": self.setup_gradient_boosting,
            "KNeighborsClassifier": self.setup_k_neighbors,
            "LinearSVC": self.setup_linear_svc,
            "LogisticRegression": self.setup_logistic_regression,
        }
        self.cs_estimators_name = self.config_space.meta["estimators"]

    @property
    def shared_hyperparameters(self):
        return {
            "alpha": [1e-3, 1e-2, 1e-1, 1.0, 10.0, 100.0],
            "fit_prior": [True, False],
            "criterion": ["gini", "entropy"],
            "max_depth": {"lower": 1, "upper": 11},
            "min_samples_split": {"lower": 2, "upper": 21},
            "min_samples_leaf": {"lower": 1, "upper": 21},
            "max_features": {"lower": 0.05, "upper": 1.01, "default_value": 1.0},
            "n_estimators": [100],
            "bootstrap": [True, False],
            "dual": [True, False],
            "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        }

    def setup_classifiers(self):
        classifiers_choices = list(self.classifiers_setup_map.keys())

        if not classifiers_choices:
            raise ValueError("No classifiers to add to config space")

        classifiers = csh.CategoricalHyperparameter(
            name=self.cs_estimators_name,
            choices=classifiers_choices,
            default_value=classifiers_choices[0],
        )
        self.config_space.add_hyperparameter(classifiers)

        for classifier_name in classifiers_choices:
            if setup_func := self.classifiers_setup_map.get(classifier_name):
                setup_func(classifiers)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, estimator_name: str
    ):
        if "classifiers" not in local_vars or not isinstance(
            local_vars["classifiers"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'classifiers' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["classifiers"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["classifiers"], estimator_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_bernoulliNB(self, classifiers: csh.CategoricalHyperparameter):
        alpha_NB = csh.CategoricalHyperparameter(
            "alpha__BernoulliNB", self.shared_hyperparameters["alpha"]
        )
        fit_prior = csh.CategoricalHyperparameter(
            "fit_prior__BernoulliNB", self.shared_hyperparameters["fit_prior"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "BernoulliNB")

    def setup_multinomialNB(self, classifiers: csh.CategoricalHyperparameter):
        alpha_NB = csh.CategoricalHyperparameter(
            "alpha__MultinomialNB", self.shared_hyperparameters["alpha"]
        )
        fit_prior = csh.CategoricalHyperparameter(
            "fit_prior__MultinomialNB", self.shared_hyperparameters["fit_prior"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "MultinomialNB")

    def setup_gaussianNB(self, classifiers: csh.CategoricalHyperparameter):
        # GaussianNB has no hyperparameters
        pass

    def setup_decision_tree(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__DecisionTreeClassifier",
            self.shared_hyperparameters["criterion"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__DecisionTreeClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "DecisionTreeClassifier"
        )

    def setup_extra_trees(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__ExtraTreesClassifier", self.shared_hyperparameters["criterion"]
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__ExtraTreesClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__ExtraTreesClassifier",
            **self.shared_hyperparameters["max_features"],
        )
        n_estimators = csh.CategoricalHyperparameter(
            "n_estimators__ExtraTreesClassifier",
            self.shared_hyperparameters["n_estimators"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__ExtraTreesClassifier", self.shared_hyperparameters["bootstrap"]
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "ExtraTreesClassifier"
        )

    def setup_random_forest(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__RandomForestClassifier",
            self.shared_hyperparameters["criterion"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__RandomForestClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split", **self.shared_hyperparameters["min_samples_split"]
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf", **self.shared_hyperparameters["min_samples_leaf"]
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features", **self.shared_hyperparameters["max_features"]
        )
        n_estimators = csh.CategoricalHyperparameter(
            "n_estimators__RandomForestClassifier",
            self.shared_hyperparameters["n_estimators"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap", self.shared_hyperparameters["bootstrap"]
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "RandomForestClassifier"
        )

    def setup_gradient_boosting(self, classifiers: csh.CategoricalHyperparameter):
        sub_sample = csh.CategoricalHyperparameter(
            "subsample", [0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        )
        learning_rate = csh.CategoricalHyperparameter(
            "learning_rate", [1e-3, 1e-2, 1e-1, 0.5, 1.0]
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__GradientBoostingClassifier",
            **self.shared_hyperparameters["max_features"],
        )
        n_estimators = csh.CategoricalHyperparameter(
            "n_estimators__GradientBoostingClassifier",
            self.shared_hyperparameters["n_estimators"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__GradientBoostingClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "GradientBoostingClassifier"
        )

    def setup_k_neighbors(self, classifiers: csh.CategoricalHyperparameter):
        n_neighbors = csh.UniformIntegerHyperparameter("n_neighbors", 1, 51)
        weights = csh.CategoricalHyperparameter("weights", ["uniform", "distance"])
        p = csh.UniformIntegerHyperparameter("p", 1, 2)
        self._add_hyperparameters_and_equals_conditions(
            locals(), "KNeighborsClassifier"
        )

    def setup_linear_svc(self, classifiers: csh.CategoricalHyperparameter):
        loss = csh.CategoricalHyperparameter(
            "loss__LinearSVC", ["hinge", "squared_hinge"]
        )
        penalty = csh.CategoricalHyperparameter("penalty__LinearSVC", ["l1", "l2"])
        dual = csh.CategoricalHyperparameter(
            "dual__LinearSVC", self.shared_hyperparameters["dual"]
        )
        tol = csh.CategoricalHyperparameter(
            "tol__LinearSVC", [1e-5, 1e-4, 1e-3, 1e-2, 1e-1]
        )
        C = csh.CategoricalHyperparameter(
            "C__LinearSVC", self.shared_hyperparameters["C"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "LinearSVC")

        # Forbidden clause: Penalty 'l1' cannot be used with loss 'hinge'
        forbidden_penalty_loss = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l1"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
        )
        # Forbidden clause: Penalty 'l2' cannot be used
        # with loss 'hinge' and dual 'False'
        forbidden_penalty_loss_dual_false = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l2"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
            cs.ForbiddenEqualsClause(self.config_space["dual__LinearSVC"], False),
        )
        self.config_space.add_forbidden_clauses(
            [forbidden_penalty_loss, forbidden_penalty_loss_dual_false]
        )

    def setup_logistic_regression(self, classifiers: csh.CategoricalHyperparameter):
        penalty = csh.CategoricalHyperparameter(
            "penalty__LogisticRegression", ["l1", "l2"]
        )
        C = csh.CategoricalHyperparameter(
            "C__LogisticRegression", self.shared_hyperparameters["C"]
        )
        dual = csh.CategoricalHyperparameter(
            "dual__LogisticRegression", self.shared_hyperparameters["dual"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "LogisticRegression")



from typing import Optional

from gama.genetic_programming.components import Individual, PrimitiveNode
from gama.genetic_programming.mutation import has_multiple_options
import ConfigSpace as cs

from gama.utilities.config_space import get_internal_longitudinal_output_types


class LongitudinalSearchSpaceConstraint:
    """LongitudinalSearchSpaceConstraint manage and apply constraints for
    genetic programming in a longitudinal context.

    This class handles the exclusion of specific preprocessors and estimators based on
    the data preparation method used. It allows for dynamic determination of valid
    mutations and crossovers within the genetic programming process.

    The constraints dict should be of the following format:
    {
        "DataPrepMethod1": {
            "preprocessors": ["Preprocessor1", "Preprocessor2", ...],
            "estimators": ["Estimator1", "Estimator2", ...]
        },
        "DataPrepMethod2": {
            "preprocessors": ["Preprocessor3", "Preprocessor4", ...],
            "estimators": ["Estimator3", "Estimator4", ...]
        },
        ...
        "Default": {
            "preprocessors": ["DefaultPreprocessor1", ...],
            "estimators": ["DefaultEstimator1", ...]
        }
    }

    How to read this format is as follows:
    - Each top-level key (e.g., "DataPrepMethod1") represents a specific data
    preparation method.
    - Under each data preparation method, there are keys for "preprocessors"
    and "estimators".
    - Each of these keys maps to a list of names of preprocessors or estimators that are
      excluded from being selected when the respective data preparation method is used.
    - The "Default" key provides fallback exclusions for any data preparation method
    not explicitly listed.

    Attributes
    __________
        constraints (dict): A dictionary defining the constraints for each data
        preparation method.

    Raises
    ______
        ValueError: If the provided constraints dictionary is invalid or missing
        required keys.
    """

    def __init__(self, constraints: dict):
        if not constraints or not isinstance(constraints, dict):
            raise ValueError(
                "No constraints provided."
                "Please provide valid constraints under a Dict format."
            )

        required_keys = {"preprocessors", "estimators"}
        for data_prep_method, constraint in constraints.items():
            if not isinstance(constraint, dict) or required_keys - constraint.keys():
                missing_keys = required_keys - constraint.keys()
                missing_keys_str = ", ".join(missing_keys)
                raise ValueError(
                    f"Constraint for '{data_prep_method}' "
                    f"must be a dictionary with keys {required_keys}. "
                    f"Missing keys: {missing_keys_str}"
                )

        self.constraints = constraints

    def get_exclusions_for_node(self, individual, node_type):
        """Determines the list of excluded options for a given node type

        Focus: of an individual

        Args
        ____
            individual (Individual): The individual being mutated.
            node_type (str): The type of node for which to determine exclusions.
            Should be one of 'data_preparation', 'preprocessors', or 'estimators'.

        Returns
        _______
            list: A list of strings representing the excluded options for the specified
            node type.
        """
        data_prep_str = self.find_data_preparation_node(individual)._primitive.__str__()

        constraints_for_method = self.constraints.get(
            data_prep_str, self.constraints.get("Default", {})
        )
        return constraints_for_method.get(node_type, [])

    def can_crossover(self, ind1_primitive, ind2_primitive):
        """Determines if two individuals can crossover

        Focus: based on their data preparation methods.

        One specific case is when the data preparation method is MerWavTimePlus, in
        which case the two individuals can only crossover if they have the same data
        preparation method.

        Args
        ____
            ind1_primitive (PrimitiveNode): The data preparation primitive of the
            first individual.
            ind2_primitive (PrimitiveNode): The data preparation primitive of the
            second individual.

        Returns
        _______
            bool: True if crossover is allowed between the two individuals,
            False otherwise.
        """
        ind1_data_preparation_primitive = self.find_data_preparation_node(
            ind1_primitive
        )._primitive
        ind2_data_preparation_primitive = self.find_data_preparation_node(
            ind2_primitive
        )._primitive

        if "MerWavTimePlus" in [
            ind1_data_preparation_primitive.__str__(),
            ind2_data_preparation_primitive.__str__(),
        ]:
            return ind1_data_preparation_primitive == ind2_data_preparation_primitive
        else:
            return True

    @staticmethod
    def find_data_preparation_node(individual: Individual):
        """Finds the data preparation node in a given individual.

        Traverses the structure of an individual to locate the node responsible for
        data preparation.

        Args
        ____
            individual (Individual): The individual to search within.

        Returns
        _______
            PrimitiveNode: The data preparation node found in the individual.

        Raises
        ______
            ValueError: If the individual does not have a data preparation node.
        """
        data_preparation_node = individual.main_node
        while data_preparation_node._primitive.output != "data_preparation":
            data_preparation_node = data_preparation_node._data_node  # type: ignore
            if not isinstance(data_preparation_node, PrimitiveNode):
                raise ValueError("Individual has no data_preparation node")
        return data_preparation_node

    @staticmethod
    def has_longitudinal_multiple_options(
        hyperparameter: cs.hyperparameters.hyperparameter,
        optional_hyperparameter_choice_name: Optional[str] = None,
    ) -> bool:
        """Checks if a ConfigSpace hyperparameter has more than one longitudinal option.

        One specific case is when the data preparation method is MerWavTimePlus, in
        which case the hyperparameter has only one option, so no multiple options.

        Args
        ____
            hyperparameter (cs.hyperparameters.hyperparameter): The hyperparameter to
            check.
            optional_hyperparameter_choice_name (str, optional): An optional
            hyperparameter choice name to consider. Defaults to None.

        Returns
        _______
            bool: True if the hyperparameter has more than one longitudinal option,
            False otherwise.
        """
        if isinstance(hyperparameter, cs.CategoricalHyperparameter):
            if "MerWavTimePlus" in hyperparameter.choices:
                if (
                    isinstance(optional_hyperparameter_choice_name, str)
                    and "MerWavTimePlus" in optional_hyperparameter_choice_name
                ):
                    return False
                return len(set(hyperparameter.choices)) > 2
            return len(set(hyperparameter.choices)) > 1
        return has_multiple_options(hyperparameter)

    @staticmethod
    def primitive_replaceable(index_primitive, config_space):
        """Determines if a primitive in an individual is replaceable

        Focus: based on longitudinal options

        One specific case is when the primitive output is a preprocessor, in which case
        the primitive is not replaceable given that the search space for preprocessors
        is currently limited to one option per preprocessor.

        Args
        ____
            index_primitive (tuple): A tuple containing the index and the
            primitive node.
            config_space (cs.ConfigurationSpace): The configuration space to
            consider for replacement options.

        Returns
        _______
            bool: True if the primitive is replaceable, False otherwise.
        """
        _, primitive = index_primitive
        if primitive._primitive.output == "preprocessors":
            return False
        return LongitudinalSearchSpaceConstraint.has_longitudinal_multiple_options(
            config_space.get_hyperparameter(
                config_space.meta[primitive._primitive.output]
                if primitive._primitive.output
                in get_internal_longitudinal_output_types()
                else primitive._primitive.output
            ),
            primitive.__str__(),
        )



from enum import Enum
from typing import Iterable, Tuple, Union

import scikit_longitudinal
from sklearn.metrics import get_scorer

# from sklearn.metrics._scorer import _ProbaScorer, _BaseScorer, _SCORERS, make_scorer
from sklearn.metrics._scorer import _Scorer, _BaseScorer, make_scorer, _SCORERS

auprc_scorer = make_scorer(scikit_longitudinal.auprc_score, needs_proba=True)
_SCORERS["auprc"] = auprc_scorer

classification_metrics = {
    "accuracy",
    "roc_auc",
    "average_precision",
    "neg_log_loss",
    "auprc",
}
for metric in ["precision", "recall", "f1"]:
    for average in ["macro", "micro", "samples", "weighted"]:
        classification_metrics.add(f"{metric}_{average}")

regression_metrics = {
    "explained_variance",
    "r2",
    "neg_mean_absolute_error",
    "neg_mean_squared_log_error",
    "neg_median_absolute_error",
    "neg_mean_squared_error",
}

all_metrics = {*classification_metrics, *regression_metrics}
reversed_scorers = {repr(v): k for k, v in _SCORERS.items()}


class MetricType(Enum):
    """Metric types supported by GAMA."""

    CLASSIFICATION: int = 1  #: discrete target
    REGRESSION: int = 2  #: continuous target


class Metric:
    """A thin layer around the `scorer` class of scikit-learn."""

    def __init__(self, scorer: Union[_BaseScorer, str]):
        if isinstance(scorer, str):
            scorer = get_scorer(scorer)
        if not isinstance(scorer, _BaseScorer):
            raise ValueError(
                "Scorer was not a valid scorer or could not be converted to one."
            )
        self.scorer = scorer
        self.name = reversed_scorers[repr(scorer)]
        self.requires_probabilities = (
            isinstance(scorer, _Scorer) and scorer._response_method == "predict_proba"
        ) or self.name == "roc_auc"

        if self.name in classification_metrics:
            self.task_type = MetricType.CLASSIFICATION
        elif self.name in regression_metrics:
            self.task_type = MetricType.REGRESSION
        else:
            raise ValueError(
                "Not sure which type of metric this is. Please raise an issue."
            )

        self.score = self.scorer._score_func

    def __call__(self, *args, **kwargs) -> float:
        return self.scorer(*args, **kwargs)

    def maximizable_score(self, *args, **kwargs) -> float:
        return self.scorer._sign * self.score(*args, **kwargs)


def scoring_to_metric(
    scoring: Union[str, Metric, Iterable[str], Iterable[Metric]]
) -> Tuple[Metric, ...]:
    if isinstance(scoring, str):
        return (Metric(scoring),)
    if isinstance(scoring, Metric):
        return (scoring,)

    if isinstance(scoring, Iterable) and all(
        isinstance(scorer, (Metric, str)) for scorer in scoring
    ):
        converted_metrics = [
            scorer if isinstance(scorer, Metric) else Metric(scorer)
            for scorer in scoring
        ]
        return tuple(converted_metrics)

    raise TypeError("scoring must be str, Metric or Iterable (of str or Metric).")



from enum import Enum
from typing import List, Tuple, Any, Union, Optional

import importlib


GamaPipelineTypeUnion = Union[
    "sklearn.pipeline.Pipeline",  # type: ignore # noqa: F821
    "scikit_longitudinal.pipeline.LongitudinalPipeline",  # type: ignore # noqa: F821
]


class GamaPipelineType(Enum):
    ScikitLearn = ("sklearn.pipeline", "Pipeline")
    ScikitLongitudinal = ("scikit_longitudinal.pipeline", "LongitudinalPipeline")

    def import_pipeline_class(self):
        module_name, class_name = self.value
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(f"Could not import module {module_name}") from e
        return getattr(module, class_name)

    @property
    def required_import(self) -> str:
        module_name, class_name = self.value
        if class_name != "Pipeline":
            return f"from {module_name} import {class_name} as Pipeline"
        return f"from {module_name} import {class_name}"


class GamaPipeline:
    def __new__(  # type: ignore
        cls,
        steps: List[Tuple[str, Any]],
        pipeline_type: Optional[GamaPipelineType] = None,
        *args,
        **kwargs,
    ) -> GamaPipelineTypeUnion:
        if steps is None or not steps:
            raise ValueError("Pipeline steps cannot be None or empty")

        if pipeline_type is None:
            raise ValueError("Pipeline type cannot be None")
        PipelineClass = pipeline_type.import_pipeline_class()
        return PipelineClass(steps, *args, **kwargs)



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class ClassifierConfig:
    """Manages the configuration space for classifiers in supervised learning contexts

    ClassifierConfig oversees the configuration space of classifiers utilised for a
    supervised machine learning tasks. This class facilitates the addition of
    new classifiers and the modification of existing ones in the configuration space
    via standardised methods. The ConfigSpace library is utilised to designate the
    configuration space, which enables the creation of intricate and adaptable
    configuration setups. For additional information on utilising constraints and
    various types of hyperparameters, with ConfigSpace, we refer the reader to
    ConfigSpace documentation, available at:
    https://automl.github.io/ConfigSpace/main/quickstart.html

    Add a classifier 💡
    ----------------

    1️⃣ To add a new classifier, define its setup method following the naming convention
    `setup_classifierName`. This method should:
        * Define hyperparameters specific to the classifier.
        * Use `_add_hyperparameters_and_equals_conditions` to add these
        hyperparameters to the config space with appropriate conditions.

    2️⃣ Next, your setup method need to be added to the `classifiers_setup_map` in
    the `__init__` method, where the key should be the Sci-kit learn name of your
    classifier, and the value should be pointing to your newly setup method.

    voila! 🎉 You are done! Your classifier is now added to the config space.

    How to use the shared hyperparameters 🪢
    -------------------------------------

    The shared hyperparameters are hyperparameters that are shared across multiple
    classifiers. These hyperparameters are defined in the `shared_hyperparameters`
    property. To use these hyperparameters, simply add them to the setup method of
    the classifier you are adding. For example, to add the `C` hyperparameter to the
    `LogisticRegression` classifier, add the following line to the
    `setup_logistic_regression` method:

    >>>    C = csh.CategoricalHyperparameter(
    >>>        "C__LogisticRegression", self.shared_hyperparameters["C"]
    >>>    )

    voila! 🎉 The `C` hyperparameter is now added to the LogisticRegression classifier
    with the shared value available in the `shared_hyperparameters` property.

    How to name my hyperparameters ✍️
    ------------------------------

    The hyperparameters you add to the config space should be named in the following
    format if similar hyperparameters names can be found in other classifiers:

    >>>    <hyperparameter_name>__<ClassifierName>

    For example, the `C` hyperparameter for the `LogisticRegression` classifier should
    be named `C__LogisticRegression` given that the `C` hyperparameter is also
    available in the `LinearSVC` classifier. This naming convention is used to ensure
    that the hyperparameters are added to the correct classifier in the config space.

    If your hyperparameter name is unique to your classifier, you can name it as you
    please without the need to have `__<ClassifierName>` at the end of the name.
    Nonetheless, following the naming convention would in any way not cause any issues.

    Modify an existing classifier 💅
    -------------------

    To modify an existing classifier, adjust its respective setup method and the
    shared hyperparameters property as needed by modifying the values of the
    hyperparameters. For example, to change the value of the `C` hyperparameter for
    the `LogisticRegression` classifier, change the value of the `C` hyperparameter
    in the `shared_hyperparameters` property by:

    >>>    "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],

    The `C` hyperparameter will then be added to the config space with the appropriate
    value. However, be cautious, if you change values in the shared hyperparameters
    property, it will be changed for all classifiers that use that hyperparameter.
    If you want this change to only apply to a specific classifier, you should add
    the hyperparameter to the setup method of that classifier. E.g. if you want to
    change the value of the `C` hyperparameter for the `LogisticRegression` classifier,
    and only want this change to apply to the `LogisticRegression` classifier, add
    the following line to the `setup_logistic_regression` method:

    >>>    C = csh.CategoricalHyperparameter(
    >>>        "C__LogisticRegression",
    >>>        [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0]
    >>>    )

    The `C` hyperparameter will be added as-is for the `LogisticRegression` classifier
    and the value of the `C` hyperparameter for other classifiers will be as available
    in the `shared_hyperparameters` property – iff they use the `C`
    hyperparameter of the `shared_hyperparameters` property.

    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that defines the hyperparameters and their ranges for
        the classifiers.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "estimators" not in config_space.meta:
            raise ValueError("Expected 'estimators' key in meta of config_space")
        self.config_space = config_space
        self.classifiers_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "BernoulliNB": self.setup_bernoulliNB,
            "MultinomialNB": self.setup_multinomialNB,
            "GaussianNB": self.setup_gaussianNB,
            "DecisionTreeClassifier": self.setup_decision_tree,
            "ExtraTreesClassifier": self.setup_extra_trees,
            "RandomForestClassifier": self.setup_random_forest,
            "GradientBoostingClassifier": self.setup_gradient_boosting,
            "KNeighborsClassifier": self.setup_k_neighbors,
            "LinearSVC": self.setup_linear_svc,
            "LogisticRegression": self.setup_logistic_regression,
        }
        self.cs_estimators_name = self.config_space.meta["estimators"]

    @property
    def shared_hyperparameters(self):
        return {
            "alpha": [1e-3, 1e-2, 1e-1, 1.0, 10.0, 100.0],
            "fit_prior": [True, False],
            "criterion": ["gini", "entropy"],
            "max_depth": {"lower": 1, "upper": 11},
            "min_samples_split": {"lower": 2, "upper": 21},
            "min_samples_leaf": {"lower": 1, "upper": 21},
            "max_features": {"lower": 0.05, "upper": 1.01, "default_value": 1.0},
            "n_estimators": {"lower": 10, "upper": 100},
            "bootstrap": [True, False],
            "dual": [True, False],
            "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        }

    def setup_classifiers(self):
        classifiers_choices = list(self.classifiers_setup_map.keys())

        if not classifiers_choices:
            raise ValueError("No classifiers to add to config space")

        classifiers = csh.CategoricalHyperparameter(
            name=self.cs_estimators_name,
            choices=classifiers_choices,
            default_value=classifiers_choices[0],
        )
        self.config_space.add_hyperparameter(classifiers)

        for classifier_name in classifiers_choices:
            if setup_func := self.classifiers_setup_map.get(classifier_name):
                setup_func(classifiers)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, estimator_name: str
    ):
        if "classifiers" not in local_vars or not isinstance(
            local_vars["classifiers"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'classifiers' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["classifiers"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["classifiers"], estimator_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_bernoulliNB(self, classifiers: csh.CategoricalHyperparameter):
        alpha_NB = csh.CategoricalHyperparameter(
            "alpha__BernoulliNB", self.shared_hyperparameters["alpha"]
        )
        fit_prior = csh.CategoricalHyperparameter(
            "fit_prior__BernoulliNB", self.shared_hyperparameters["fit_prior"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "BernoulliNB")

    def setup_multinomialNB(self, classifiers: csh.CategoricalHyperparameter):
        alpha_NB = csh.CategoricalHyperparameter(
            "alpha__MultinomialNB", self.shared_hyperparameters["alpha"]
        )
        fit_prior = csh.CategoricalHyperparameter(
            "fit_prior__MultinomialNB", self.shared_hyperparameters["fit_prior"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "MultinomialNB")

    def setup_gaussianNB(self, classifiers: csh.CategoricalHyperparameter):
        # GaussianNB has no hyperparameters
        pass

    def setup_decision_tree(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__DecisionTreeClassifier",
            self.shared_hyperparameters["criterion"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__DecisionTreeClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__DecisionTreeClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "DecisionTreeClassifier"
        )

    def setup_extra_trees(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__ExtraTreesClassifier", self.shared_hyperparameters["criterion"]
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__ExtraTreesClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__ExtraTreesClassifier",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__ExtraTreesClassifier",
            **self.shared_hyperparameters["max_features"],
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__ExtraTreesClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__ExtraTreesClassifier", self.shared_hyperparameters["bootstrap"]
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "ExtraTreesClassifier"
        )

    def setup_random_forest(self, classifiers: csh.CategoricalHyperparameter):
        criterion = csh.CategoricalHyperparameter(
            "criterion__RandomForestClassifier",
            self.shared_hyperparameters["criterion"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__RandomForestClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split", **self.shared_hyperparameters["min_samples_split"]
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf", **self.shared_hyperparameters["min_samples_leaf"]
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features", **self.shared_hyperparameters["max_features"]
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators", **self.shared_hyperparameters["n_estimators"]
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap", self.shared_hyperparameters["bootstrap"]
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "RandomForestClassifier"
        )

    def setup_gradient_boosting(self, classifiers: csh.CategoricalHyperparameter):
        sub_sample = csh.UniformFloatHyperparameter(
            "subsample", 0.05, 1.01, default_value=1.0
        )
        learning_rate = csh.CategoricalHyperparameter(
            "learning_rate", [1e-3, 1e-2, 1e-1, 0.5, 1.0]
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__GradientBoostingClassifier",
            **self.shared_hyperparameters["max_features"],
        )
        n_estimators = csh.UniformIntegerHyperparameter(
            "n_estimators__GradientBoostingClassifier",
            **self.shared_hyperparameters["n_estimators"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__GradientBoostingClassifier",
            **self.shared_hyperparameters["max_depth"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "GradientBoostingClassifier"
        )

    def setup_k_neighbors(self, classifiers: csh.CategoricalHyperparameter):
        n_neighbors = csh.UniformIntegerHyperparameter("n_neighbors", 1, 51)
        weights = csh.CategoricalHyperparameter("weights", ["uniform", "distance"])
        p = csh.UniformIntegerHyperparameter("p", 1, 2)
        self._add_hyperparameters_and_equals_conditions(
            locals(), "KNeighborsClassifier"
        )

    def setup_linear_svc(self, classifiers: csh.CategoricalHyperparameter):
        loss = csh.CategoricalHyperparameter(
            "loss__LinearSVC", ["hinge", "squared_hinge"]
        )
        penalty = csh.CategoricalHyperparameter("penalty__LinearSVC", ["l1", "l2"])
        dual = csh.CategoricalHyperparameter(
            "dual__LinearSVC", self.shared_hyperparameters["dual"]
        )
        tol = csh.CategoricalHyperparameter(
            "tol__LinearSVC", [1e-5, 1e-4, 1e-3, 1e-2, 1e-1]
        )
        C = csh.CategoricalHyperparameter(
            "C__LinearSVC", self.shared_hyperparameters["C"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "LinearSVC")

        # Forbidden clause: Penalty 'l1' cannot be used with loss 'hinge'
        forbidden_penalty_loss = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l1"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
        )
        # Forbidden clause: Penalty 'l2' cannot be used with loss
        # 'hinge' and dual 'False'
        forbidden_penalty_loss_dual_false = cs.ForbiddenAndConjunction(
            cs.ForbiddenEqualsClause(self.config_space["penalty__LinearSVC"], "l2"),
            cs.ForbiddenEqualsClause(self.config_space["loss__LinearSVC"], "hinge"),
            cs.ForbiddenEqualsClause(self.config_space["dual__LinearSVC"], False),
        )
        self.config_space.add_forbidden_clauses(
            [forbidden_penalty_loss, forbidden_penalty_loss_dual_false]
        )

    def setup_logistic_regression(self, classifiers: csh.CategoricalHyperparameter):
        solver = csh.CategoricalHyperparameter("solver", ["lbfgs"])
        penalty = csh.CategoricalHyperparameter("penalty__LogisticRegression", ["l2"])
        C = csh.CategoricalHyperparameter(
            "C__LogisticRegression", self.shared_hyperparameters["C"]
        )
        dual = csh.CategoricalHyperparameter(
            "dual__LogisticRegression", self.shared_hyperparameters["dual"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "LogisticRegression")



import ast
from typing import List, Union, cast


from .terminal import DATA_TERMINAL, Terminal
from .primitive import Primitive
import ConfigSpace as cs

from ...utilities.config_space import (
    get_hyperparameter_sklearn_name,
    get_estimator_by_name,
    get_longitudinal_estimator_by_name,
)


class PrimitiveNode:
    """An instantiation for a Primitive with specific Terminals.

    Parameters
    ----------
    primitive: Primitive
        The Primitive type of this PrimitiveNode.
    data_node: PrimitiveNode
        The PrimitiveNode that specifies all preprocessing before this PrimitiveNode.
    terminals: List[Terminal]
        A list of terminals matching the `primitive`.
    """

    def __init__(
        self,
        primitive: Primitive,
        data_node: Union["PrimitiveNode", str],
        terminals: List[Terminal],
    ):
        self._primitive = primitive
        self._data_node = data_node
        self._terminals = sorted(terminals, key=lambda t: str(t))

    def __str__(self) -> str:
        """Recursively stringify all primitive nodes (primitive and hyperparameters).

        Examples: - "GaussianNB(data)"
                  - "BernoulliNB(data, alpha=1.0)"
                  - "BernoulliNB(FastICA(data, tol=0.5), alpha=1.0)"
        """
        if not self._terminals:
            return f"{self._primitive}({self._data_node})"
        terminal_str = ", ".join([repr(terminal) for terminal in self._terminals])
        return f"{self._primitive}({self._data_node}, {terminal_str})"

    @property
    def str_nonrecursive(self) -> str:
        """Stringify primitive node with its hyperparameter configuration

        Examples: - "GaussianNB()"
                  - "BernoulliNB(alpha=1.0)"
        """
        terminal_str = ", ".join([str(terminal) for terminal in self._terminals])
        return f"{self._primitive}({terminal_str})"

    def copy(self) -> "PrimitiveNode":
        """Copies the object. Shallow for terminals, deep for data_node."""
        if isinstance(self._data_node, str) and self._data_node == DATA_TERMINAL:
            data_node_copy = DATA_TERMINAL  # type: Union[str, PrimitiveNode]
        elif isinstance(self._data_node, PrimitiveNode):
            data_node_copy = self._data_node.copy()
        else:
            raise ValueError(
                f"self._data_node is neither str nor PrimitiveNode, "
                f"but {type(self._data_node)}"
            )
        return PrimitiveNode(
            primitive=self._primitive,
            data_node=data_node_copy,
            terminals=self._terminals.copy(),
        )

    @classmethod
    def from_string(
        cls, string: str, config_space: cs.ConfigurationSpace, strict: bool = True
    ) -> "PrimitiveNode":
        """Create a PrimitiveNode from string formatted like PrimitiveNode.__str__

        Parameters
        ----------
        string: str
            A string formatted similar to PrimitiveNode.__str__
        config_space: ConfigurationSpace
            The ConfigSpace object which defines the search space. Refer to the
            configuration/(classification||regression).py file for further details.
        strict: bool (default=True)
            Require each primitives has all required terminals present in `string`.
            Non-strict matching may be useful when constructing individuals from
            and old log with a slightly different search space.

        Returns
        -------
        PrimitiveNode
            The PrimitiveNode as defined the string.
        """
        # General form is
        # A(B(C(data[, C.param=value, ..])[, B.param=value, ..])[, A.param=value, ..])
        # below assumes that left parenthesis is never part of a parameter name or value
        primitives = string.split("(")[:-1]
        terminal_start_index = string.index(DATA_TERMINAL)
        terminals_string = string[terminal_start_index + len(DATA_TERMINAL) :]
        terminal_sets = terminals_string.split(")")[:-1]

        last_node: Union[PrimitiveNode, str] = DATA_TERMINAL
        for primitive_string, terminal_set in zip(reversed(primitives), terminal_sets):
            primitive = find_primitive(config_space, primitive_string)
            if terminal_set == "":
                terminals = []
            else:
                terminal_set = terminal_set[2:]  # 2 is because string starts with ', '
                terminals = [
                    find_terminal(config_space, terminal_string, primitive_string)
                    for terminal_string in terminal_set.split(", ")
                ]
            missing = set(primitive.input) - set(map(lambda t: t.identifier, terminals))
            if missing and strict:
                raise ValueError(f"terminals {missing} for primitive {primitive}")
            last_node = cls(primitive, last_node, terminals)

        return cast(PrimitiveNode, last_node)


def find_primitive(
    config_space: cs.ConfigurationSpace, primitive_string: str
) -> Primitive:
    """Find the Primitive that matches `primitive_string` in `config_space`."""
    if config_space is None:
        raise ValueError("config_space must not be None")
    if "estimators" not in config_space.meta:
        raise ValueError(
            "config_space must have meta information about the estimators"
            "hyperparameters"
        )

    estimators = config_space.get_hyperparameter(
        config_space.meta["estimators"]
    ).choices
    preprocessors = None
    if "preprocessors" in config_space.meta:
        preprocessors = config_space.get_hyperparameter(
            config_space.meta["preprocessors"]
        ).choices

    data_preparation = None
    if "data_preparation" in config_space.meta:
        data_preparation = config_space.get_hyperparameter(
            config_space.meta["data_preparation"]
        ).choices
        data_preparation += data_preparation

    all_hyperparameters = []
    if estimators:
        all_hyperparameters += estimators
    if preprocessors:
        all_hyperparameters += preprocessors
    if data_preparation:
        all_hyperparameters += data_preparation

    if primitive_string in all_hyperparameters:
        if primitive_string in estimators:
            output = "estimators"
        elif primitive_string in preprocessors:
            output = "preprocessors"
        else:
            output = "data_preparation"
        try:
            primitive = Primitive(
                input=(),
                output=output,
                identifier=get_estimator_by_name(primitive_string),
            )
        except ValueError:
            primitive = Primitive(
                input=(),
                output=output,
                identifier=get_longitudinal_estimator_by_name(primitive_string),
            )
        return primitive

    raise IndexError(f"Could not find Primitive of type '{primitive_string}'.")


def find_terminal(
    config_space: cs.ConfigurationSpace, terminal_string: str, primitive_string: str
) -> Terminal:
    """Find the Terminal that matches `terminal_string` in `config_space`."""
    if config_space is None:
        raise ValueError("config_space must not be None")

    term_type, value = terminal_string.split("=")
    if "." in term_type:
        term_parent_type, term_type = term_type.split(".")
        term_config_space_name = f"{term_type}__{term_parent_type}"
    else:
        term_config_space_name = f"{term_type}__{primitive_string}"

    if isinstance(value, str):
        value = value.replace("'", "").replace('"', "").replace(" ", "")
        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            value = str(value)

    if term_config_space_name in config_space.get_hyperparameter_names():
        return Terminal(
            identifier=get_hyperparameter_sklearn_name(term_config_space_name),
            value=value,
            output=get_hyperparameter_sklearn_name(term_config_space_name),
            config_space_name=term_config_space_name,
        )
    if term_type in config_space.get_hyperparameter_names():
        return Terminal(
            identifier=get_hyperparameter_sklearn_name(term_type),
            value=value,
            output=get_hyperparameter_sklearn_name(term_type),
            config_space_name=term_type,
        )

    raise RuntimeError(f"Could not find Terminal of type '{terminal_string}'.")



import ConfigSpace as cs

from .classification_task import ClassifierConfig, PreprocessorConfig

# Classifiers & Preprocessors 🚀

# This script is your ticket to configuring a ConfigSpace object, teeming with
# classifiers and preprocessors. We are diving in with the ClassifierConfig and
# PreprocessorConfig classes to fill the configuration space with a slew of
# hyperparameters and options.

# Customise Your Space 🔧

# Want just classifiers? No biggie! Just comment out or remove the PreprocessorConfig
# setup + meta key-value & Voila! You're left with a sleek, classifier-only
# configuration space.

# Want to add more classifiers or preprocessors? Easy! Just add them to the
# ClassifierConfig or PreprocessorConfig classes, respectively. You can even
# add your own custom classifiers or preprocessors. Just make sure they are
# compatible with scikit-learn's API.

# Meta-Parameters 📝

# The meta-parameters are the "estimators" and "preprocessors" keys in the
# configuration space. These are used to identify the classifiers and preprocessors
# by the internal system. They are not hyperparameters, and should not be
# changed, except by advanced users. If you do change them, make sure to change
# the corresponding values in the current configuration space, i.e. in ClassifierConfig
# and PreprocessorConfig.

# 👩‍💻👨‍💻 Happy configuring, and may your machine learning models shine!

config_space = cs.ConfigurationSpace(
    meta={
        # "gama_system_name": "current_configuration_name",
        "estimators": "classifiers",
        "preprocessors": "preprocessors",
    }
)

classifier_config = ClassifierConfig(config_space)
classifier_config.setup_classifiers()

preprocessor_config = PreprocessorConfig(config_space)
preprocessor_config.setup_preprocessors()



import datetime
import heapq
import logging
from typing import Tuple, List, Optional, Union, Dict
import pickle
import os

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.model_selection import StratifiedShuffleSplit

from gama.genetic_programming.components import Individual

log = logging.getLogger(__name__)


class Evaluation:
    """Record relevant evaluation data of an individual."""

    def __init__(
        self,
        individual: Individual,
        predictions: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None,
        score: Tuple[float, ...] = (),
        estimators: Optional[List] = None,
        start_time: Optional[datetime.datetime] = None,
        duration: float = -1,
        error: Optional[str] = None,
        pid: Optional[int] = None,
    ):
        self.individual: Individual = individual
        self.score = score
        self._estimators: List[BaseEstimator] = [] if estimators is None else estimators
        self.start_time = start_time
        self.duration = duration
        self.error = error
        self.pid = pid
        self._cache_file = ""

        if isinstance(predictions, (pd.Series, pd.DataFrame)):
            predictions = predictions.values
        self._predictions: Optional[np.ndarray] = predictions

    def to_disk(self, directory: str) -> None:
        """Save Evaluation in the provided directory."""
        self._cache_file = os.path.join(directory, f"{str(self.individual._id)}.pkl")
        with open(self._cache_file, "wb") as fh:
            pickle.dump((self._estimators, self._predictions), fh)
        self._estimators, self._predictions = [], None

    def remove_from_disk(self) -> None:
        """Remove the related file from disk."""
        os.remove(os.path.join(self._cache_file))
        self._cache_file = ""

    @property
    def estimators(self) -> List[BaseEstimator]:
        if self._estimators or not self._cache_file:
            return self._estimators
        with open(self._cache_file, "rb") as fh:
            estimators, _ = pickle.load(fh)
            return estimators

    @property
    def predictions(self):
        if self._predictions is not None or not self._cache_file:
            return self._predictions
        with open(self._cache_file, "rb") as fh:
            _, predictions = pickle.load(fh)
            return predictions

    # Is there a better way to do this?
    # Assignment in __init__ is not preferred even if it saves lines.
    def __lt__(self, other):
        return self.score.__lt__(other.score)

    def __le__(self, other):
        return self.score.__le__(other.score)

    def __eq__(self, other):
        return self.score.__eq__(other.score)

    def __ne__(self, other):
        return self.score.__ne__(other.score)

    def __gt__(self, other):
        return self.score.__gt__(other.score)

    def __ge__(self, other):
        return self.score.__ge__(other.score)


class EvaluationLibrary:
    """Maintains an in-memory record of evaluations.

    The main function of the EvaluationLibrary is to maintain a fast lookup for
    the best evaluations, and to discard meta-data of Evaluations which are not
    good enough.

    Specifically for the top `m` evaluations the estimators and predictions are kept.
    All but `n` predictions of each evaluation are discarded.
    As soon as an evaluation is no longer in the top `m`,
    its estimators and the sampled predictions are also discarded.
    Other evaluation meta-data (e.g. scores, evaluation time, errors) is not discarded.

    This discarding is useful to reduce memory usage when you know which meta-data
    is used later. E.g.:
    Ensembling selects from the best `m` models and uses `n` points for hill-climbing,
    but general score information is relevant for all evaluated models.
    """

    def __init__(
        self,
        m: Optional[int] = 200,
        n: Optional[int] = 10_000,
        sample: Optional[np.ndarray] = None,
        cache: str = "cache",
    ):
        """Create an EvaluationLibrary for in-memory record of evaluations.

        Parameters
        ----------
        m: int, optional (default=200)
            Evaluations outside of the top `m` have their predictions and estimators
            discarded to reduce memory usage.
            The best `m` evaluations can be queried quickly through `n_best(int)`.
            If `None`, never discard predictions and estimators.
        n: int, optional (default=10_000)
            Instead of storing all predictions of the 'top m' evaluations,
            store only 'n' predictions. If left `None` all predictions will be kept.
            This parameter is ignored when 'sample' is also provided.
            Call `determine_sample_indices` before saving the first evaluation,
            if you want to ensure a class stratified sample.
        sample: np.ndarray, optional (default=None)
            Instead of storing all predictions of the 'top m' evaluations,
            store only those with indices specified in this array.
        cache: str (default="cache")
            Directory to save evaluations to.
            For large datasets or a big number of models,
            this can lead to memory issues.
        """
        self.top_evaluations: List[Evaluation] = []
        self.other_evaluations: List[Evaluation] = []
        self._m = m
        self._sample_n = n
        self.lookup: Dict[str, Evaluation] = {}
        self._cache = os.path.expandvars(cache)
        if not os.path.exists(self._cache):
            os.mkdir(self._cache)

        def main_node_str(e: Evaluation):
            return str(e.individual.main_node)

        self._lookup_key = main_node_str

        if sample is not None:
            self._sample = sample
        elif n is None:
            self._sample = None
        else:
            self._sample = "not set"

    @property
    def evaluations(self) -> List[Evaluation]:
        return self.top_evaluations + self.other_evaluations

    def determine_sample_indices(
        self,
        n: Optional[int] = None,
        prediction_size: Optional[int] = None,
        stratify: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None,
    ) -> None:
        """Set `self._sample` to an array for sampling predictions or `None`.

        The sample indices can be class stratified if `stratify` is set.
        If `prediction_size` or `len(stratify)` is smaller than `n`,
        predictions will not be sampled and all predictions are saved.

        Parameters
        ----------
        n: int, optional (default=None)
            Number of predictions to keep for each evaluation.
            If `None`, use `n` set at the library's initialisation.
        prediction_size: int, optional (default=None)
            Number of predictions one can sample from.
        stratify: np.ndarray, pd.Series or pd.DataFrame, optional (default=None)
            The target variable of a supervised classification problem.
            The sample will be class stratified according to this variable.
        """
        if not ((prediction_size is None) ^ (stratify is None)):
            raise ValueError(
                "Exactly one of `prediction_size` and `stratify` must be set."
            )
        if len(self.top_evaluations) > 0:
            log.warning("New subsample not used for already stored evaluations.")
        n = self._sample_n if n is None else n

        if n is None:
            # No n was provided here nor set on initialization
            self._sample = None
        elif prediction_size is not None and n < prediction_size:
            # Subsample is to be chosen uniformly random.
            self._sample = np.random.choice(
                range(prediction_size), size=n, replace=False
            )
        elif stratify is not None and n < len(stratify):
            splitter = StratifiedShuffleSplit(n_splits=1, train_size=n)
            self._sample, _ = next(splitter.split(np.zeros(len(stratify)), stratify))
        else:
            # Specified sample size exceeds size of predictions
            self._sample = None

    def _process_predictions(self, evaluation: Evaluation) -> None:
        """Downsample evaluation predictions if required."""
        if self._sample_n == 0:
            evaluation._predictions = None
        if evaluation.predictions is None:
            return  # Predictions either not provided or removed because sample_n is 0.

        if isinstance(self._sample, str) and self._sample == "not set":
            # Happens only for the first evaluation with predictions.
            self.determine_sample_indices(self._sample_n, len(evaluation.predictions))

        if self._sample is not None:
            evaluation._predictions = evaluation.predictions[self._sample]

    def save_evaluation(self, evaluation: Evaluation) -> None:
        self._process_predictions(evaluation)

        if evaluation.error is not None:
            evaluation._estimators, evaluation._predictions = [], None
            self.other_evaluations.append(evaluation)
        elif self._m is None or self._m > len(self.top_evaluations):
            evaluation.to_disk(self._cache)
            heapq.heappush(self.top_evaluations, evaluation)
        else:
            removed = heapq.heappushpop(self.top_evaluations, evaluation)
            if removed == evaluation:
                # new evaluation is not in heap, big memory items may be discarded
                removed._estimators, removed._predictions = [], None
            else:
                # new evaluation is now on the heap, remove old from disk
                evaluation.to_disk(self._cache)
                removed.remove_from_disk()

            self.other_evaluations.append(removed)

        self.lookup[self._lookup_key(evaluation)] = evaluation

    def clear_cache(self) -> None:
        for file in os.listdir(self._cache):
            os.remove(os.path.join(self._cache, file))
        os.rmdir(self._cache)

    def n_best(self, n: int = 5, with_pipelines=True) -> List[Evaluation]:
        """Return the best `n` pipelines.

        Slower if `n` exceeds `m` given on initialization.
        """
        if self._m is None or n <= self._m or with_pipelines:
            return heapq.nlargest(n, self.top_evaluations)
        else:
            return list(reversed(sorted(self.evaluations)))[:n]

    def is_evaluated(self, candidate: Union[Individual, None]) -> bool:
        """Check if a candidate pipeline has already been evaluated."""
        if candidate is None:
            log.warning("Candidate to check is None. Returning False.")
            return False
        return any(
            str(candidate.pipeline) == str(evaluation.individual.pipeline)
            for evaluation in self.lookup.values()
        )



import inspect
from typing import Union, Optional
import logging

import numpy as np
import pandas as pd
from ConfigSpace import ForbiddenEqualsClause
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder
import ConfigSpace as cs

from gama.configuration.classification import config_space as clf_config
from gama.data_loading import X_y_from_file
from gama.utilities.metrics import scoring_to_metric, Metric
from .gama import Gama
from .utilities.config_space import get_estimator_by_name

# Avoid stopit from logging warnings every time a pipeline evaluation times out
logging.getLogger("stopit").setLevel(logging.ERROR)
log = logging.getLogger(__name__)


class GamaClassifier(Gama):
    """Gama with adaptations for (multi-class) classification."""

    def __init__(
        self,
        search_space: Optional[cs.ConfigurationSpace] = None,
        scoring: Metric = "neg_log_loss",  # type: ignore
        *args,
        **kwargs,
    ):
        if not search_space:
            # Do this to avoid the whole dictionary being included in the documentation.
            search_space = clf_config

        self._metrics = scoring_to_metric(scoring)

        search_space = self._search_space_check(search_space)

        self._label_encoder = None
        super().__init__(
            *args, search_space=search_space, scoring=scoring, **kwargs
        )  # type: ignore

    def _search_space_check(
        self,
        search_space: cs.ConfigurationSpace,
    ) -> cs.ConfigurationSpace:
        """Check if the search space is valid for classification."""

        # Check if the search space contains a classifier hyperparameter.
        if (
            "estimators" not in search_space.meta
            or (
                search_space.meta["estimators"]
                not in search_space.get_hyperparameters_dict()
            )
            or not isinstance(
                search_space.get_hyperparameter(search_space.meta["estimators"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the classifiers "
                "that is a CategoricalHyperparameter with choices for all desired "
                "classifiers. Please double-check the spelling of the name, and review "
                "the `meta` object in the search space configuration located at "
                "`configurations/classification.py`. The `meta` object should contain "
                "a key `estimators` with a value that is the name of the hyperparameter"
                " that contains the classifier choices."
            )

        # Check if the search space contains a preprocessor hyperparameter
        # if it is specified in the meta.
        if (
            "preprocessors" in search_space.meta
            and (
                search_space.meta["preprocessors"]
                not in search_space.get_hyperparameters_dict()
            )
            or "preprocessors" in search_space.meta
            and not isinstance(
                search_space.get_hyperparameter(search_space.meta["preprocessors"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the preprocessors "
                "that is a CategoricalHyperparameter with choices for all desired "
                "preprocessors. Please double-check the spelling of the name, and "
                "review the `meta` object in the search space configuration located at "
                "`configurations/classification.py`. The `meta` object should contain "
                "a key `preprocessors` with a value that is the name of the "
                "hyperparameter that contains the preprocessor choices. "
            )

        # Check if the search space contains only classifiers that have predict_proba
        # if the scoring requires probabilities.
        if any(metric.requires_probabilities for metric in self._metrics):
            # we don't want classifiers that do not have `predict_proba`,
            # because then we have to start doing one hot encodings of predictions etc.
            no_proba_clfs = []
            for classifier in search_space.get_hyperparameter(
                search_space.meta["estimators"]
            ).choices:
                if classifier == "Dummy_To_Ignore":
                    continue
                estimator = get_estimator_by_name(classifier)
                if (
                    estimator is not None
                    and issubclass(estimator, ClassifierMixin)
                    and not hasattr(estimator(), "predict_proba")
                ):
                    no_proba_clfs.append(classifier)

            log.info(
                f"The following classifiers do not have a predict_proba method "
                f"and will be excluded from the search space: {no_proba_clfs}"
            )
            search_space.add_forbidden_clauses(
                [
                    ForbiddenEqualsClause(
                        search_space.get_hyperparameter(
                            search_space.meta["estimators"]
                        ),
                        classifier,
                    )
                    for classifier in no_proba_clfs
                    if classifier
                ]
            )

        return search_space

    def _predict(self, x: pd.DataFrame):
        """Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array with predictions of shape (N,) where N is len(X).
        """
        y = self.model.predict(x)  # type: ignore
        # Decode the predicted labels - necessary only if ensemble is not used.
        if y[0] not in list(self._label_encoder.classes_):  # type: ignore
            y = self._label_encoder.inverse_transform(y)  # type: ignore
        return y

    def _predict_proba(self, x: pd.DataFrame):
        """Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        return self.model.predict_proba(x)  # type: ignore

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        """Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            Data with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def predict_proba_from_file(
        self,
        arff_file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
    ):
        """Predict the class probabilities for input in the arff_file.

        Parameters
        ----------
        arff_file_path: str
            An ARFF file with the same columns as the one that used in fit.
            Target column must be present in file, but its values are ignored.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the ARFF file.

        Returns
        -------
        numpy.ndarray
            Numpy array with class probabilities.
            The array is of shape (N, K) where N is len(X),
            and K is the number of class labels found in `y` of `fit`.
        """
        x, _ = X_y_from_file(arff_file_path, target_column, encoding)
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def fit(self, x, y, *args, **kwargs):
        """Should use base class documentation."""
        y_ = y.squeeze() if isinstance(y, pd.DataFrame) else y
        self._label_encoder = LabelEncoder().fit(y_)
        if any(isinstance(yi, str) for yi in y_):
            # If target values are `str` we encode them or scikit-learn will complain.
            y = self._label_encoder.transform(y_)
        self._evaluation_library.determine_sample_indices(stratify=y)

        # Add label information for classification to the scorer such that
        # the cross validator does not encounter unseen labels in smaller
        # data sets during pipeline evaluation.
        for m in self._metrics:
            if "labels" in inspect.signature(m.scorer._score_func).parameters:
                m.scorer._kwargs.update({"labels": y})

        super().fit(x, y, *args, **kwargs)

    def _encode_labels(self, y):
        self._label_encoder = LabelEncoder().fit(y)
        return self._label_encoder.transform(y)



import logging
from typing import Optional, Iterator, List, Tuple
import category_encoders as ce
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.impute import SimpleImputer
from gama.GamaPipeline import GamaPipeline, GamaPipelineTypeUnion, GamaPipelineType

log = logging.getLogger(__name__)


def select_categorical_columns(
    df: pd.DataFrame,
    min_f: Optional[int] = None,
    max_f: Optional[int] = None,
    ignore_nan: bool = True,
) -> Iterator[str]:
    """Find all categorical columns with at least `min_f` and at most `max_f` factors.

    Parameters
    ----------
    df: pd.DataFrame
        Pandas DataFrame to design the encoder for.
    min_f: int, optional (default=None)
        The inclusive minimum number of unique values the column should have.
    max_f: int, optional (default=None)
        The inclusive maximum number of unique values the column should have.
    ignore_nan: bool (default=True)
        If True, don't count NaN as a unique value.
        If False, count NaN as a unique value (only once).

    Returns
    -------
    An iterator which iterates over the column names that satisfy the criteria.
    """
    for column in df:
        if isinstance(df[column].dtype, pd.core.dtypes.dtypes.CategoricalDtype):
            nfactors = df[column].nunique(dropna=ignore_nan)
            if (min_f is None or min_f <= nfactors) and (
                max_f is None or nfactors <= max_f
            ):
                yield column


def basic_encoding(
    x: pd.DataFrame,
    is_classification: bool,
    gama_pipeline_type: Optional[GamaPipelineTypeUnion] = GamaPipelineType.ScikitLearn,
) -> Tuple[pd.DataFrame, GamaPipelineTypeUnion]:
    """Perform 'basic' encoding of categorical features.

    Specifically, perform:
     - Ordinal encoding for features with 2 or fewer unique values.
     - One hot encoding for features with at most 10 unique values.
     - Ordinal encoding for features with 11+ unique values, if y is categorical.
    """
    ord_features = list(select_categorical_columns(x, max_f=2))
    if is_classification:
        ord_features.extend(select_categorical_columns(x, min_f=11))
    leq_10_features = list(select_categorical_columns(x, min_f=3, max_f=10))

    encoding_steps = [
        ("ord-enc", ce.OrdinalEncoder(cols=ord_features, drop_invariant=True)),
        ("oh-enc", ce.OneHotEncoder(cols=leq_10_features, handle_missing="value")),
    ]
    if gama_pipeline_type:
        encoding_pipeline = GamaPipeline(
            encoding_steps, pipeline_type=gama_pipeline_type
        )
    else:
        encoding_pipeline = GamaPipeline(encoding_steps)
    if not hasattr(encoding_pipeline, "fit_transform"):
        raise ValueError(
            f"Pipeline {encoding_pipeline} does not have a fit_transform method."
        )
    x_enc = encoding_pipeline.fit_transform(x, y=None)
    return x_enc, encoding_pipeline


def basic_pipeline_extension(
    x: pd.DataFrame, is_classification: bool
) -> List[Tuple[str, TransformerMixin]]:
    """Define a TargetEncoder and SimpleImputer.

    TargetEncoding is will encode categorical features with more than 10 unique values,
    if y is not categorical. SimpleImputer imputes with the median.
    """
    # These steps need to be in the pipeline because they need to be trained each fold.
    extension_steps = []
    if not is_classification:
        # TargetEncoder is broken with categorical target
        many_factor_features = list(select_categorical_columns(x, min_f=11))
        extension_steps.append(
            ("target_enc", ce.TargetEncoder(cols=many_factor_features))
        )
    extension_steps.append(("imputation", SimpleImputer(strategy="median")))

    return extension_steps



import logging
from typing import List, Optional

import pandas as pd

from gama.genetic_programming.components import Individual
from gama.genetic_programming.operator_set import OperatorSet
from gama.search_methods.base_search import (
    BaseSearch,
    _check_base_search_hyperparameters,
)
from gama.utilities.generic.async_evaluator import AsyncEvaluator

log = logging.getLogger(__name__)


class RandomSearch(BaseSearch):
    """Perform random search over all possible pipelines."""

    def dynamic_defaults(
        self, x: pd.DataFrame, y: pd.DataFrame, time_limit: float
    ) -> None:
        pass

    def search(
        self, operations: OperatorSet, start_candidates: List[Individual]
    ) -> None:
        random_search(operations, self.output, start_candidates)


def random_search(
    operations: OperatorSet,
    output: List[Individual],
    start_candidates: List[Individual],
    max_evaluations: Optional[int] = None,
    max_attempts: int = 100000,
) -> List[Individual]:
    """Perform random search over all possible pipelines

    Parameters
    ----------
    operations: OperatorSet
        An operator set with `evaluate` and `individual` functions.
    output: List[Individual]
        A list which contains the found individuals during search.
    start_candidates: List[Individual]
        A list with candidate individuals to evaluate first.
    max_evaluations: int, optional (default=None)
        If specified, only a maximum of `max_evaluations` individuals are evaluated.
        If None, the algorithm will be run indefinitely.
    max_attempts: int (default=100000)
        Maximum number of attempts to generate a unique individual otherwise raise
        an error.

    Returns
    -------
    List[Individual]
        All evaluated individuals.
    """
    _check_base_search_hyperparameters(operations, output, start_candidates)

    with AsyncEvaluator() as async_:
        for individual in start_candidates:
            async_.submit(operations.evaluate, individual)

        while (max_evaluations is None) or (len(output) < max_evaluations):
            future = operations.wait_next(async_)
            if future.result is not None:
                output.append(future.result.individual)

            attempts = 0
            while (
                new_individual := operations.individual()
            ) and operations.is_evaluated(
                new_individual
            ):  # type: ignore
                if attempts >= max_attempts:
                    raise ValueError(
                        "Maximum attempts reached while trying to generate a"
                        "unique individual."
                    )
                attempts += 1

            async_.submit(operations.evaluate, new_individual)

    return output
